#!/usr/bin/env python3
# encoding: utf-8
import sys
import logging
from datetime import datetime
import pytest
import math
import os
from pathlib import Path
import itertools

import leuvenmapmatching as mm
from leuvenmapmatching.util import dist_euclidean as de
from leuvenmapmatching.util import dist_latlon as dll


directory = None


def test_path_to_gpx():
    from leuvenmapmatching.util.gpx import path_to_gpx
    path = [(i, i, datetime.fromtimestamp(i)) for i in range(0, 10)]
    gpx = path_to_gpx(path)

    assert len(path) == len(gpx.tracks[0].segments[0].points)
    assert path[0][0] == pytest.approx(gpx.tracks[0].segments[0].points[0].latitude)
    assert path[0][1] == pytest.approx(gpx.tracks[0].segments[0].points[0].longitude)
    assert path[0][2] == gpx.tracks[0].segments[0].points[0].time


def test_grs80():
    from leuvenmapmatching.util.projections import latlon2grs80
    coordinates = [(4.67878, 50.864), (4.68054, 50.86381), (4.68098, 50.86332), (4.68129, 50.86303), (4.6817, 50.86284),
                   (4.68277, 50.86371), (4.68894, 50.86895), (4.69344, 50.86987), (4.69354, 50.86992),
                   (4.69427, 50.87157), (4.69643, 50.87315), (4.69768, 50.87552), (4.6997, 50.87828)]
    points = latlon2grs80(coordinates, lon_0=coordinates[0][0], lat_ts=coordinates[0][1])
    points = list(points)
    point = points[0]
    assert point[0] == pytest.approx(618139.9385518166)
    assert point[1] == pytest.approx(5636043.991970774)


def test_distance1():
    p1 = (38.898556, -77.037852)
    p2 = (38.897147, -77.043934)
    d = dll.distance(p1, p2)
    assert d == pytest.approx(549.1557912048178), f"Got: {d}"


def test_distance2():
    o_p1 = (6007539.987516373, -13607675.997610645)
    m_p1 = (6007518.475594072, -13607641.049711559)
    m_p2 = (6007576.295597112, -13607713.306589901)
    dist, proj_m, t_m = de.distance_point_to_segment(o_p1, m_p1, m_p2)
    assert dist == pytest.approx(5.038773480896327), f"dist = {dist}"
    assert t_m == pytest.approx(0.4400926470800718), f"t_m = {t_m}"


def test_bearing1():
    lat1, lon1 = math.radians(38.898556), math.radians(-77.037852)
    lat2, lon2 = math.radians(38.897147), math.radians(-77.043934)
    b = dll.bearing_radians(lat1, lon1, lat2, lon2)
    b = math.degrees(b)
    # assert b == pytest.approx(253.42138889), f"Got: {b}"
    assert b == pytest.approx(-106.5748183426045), f"Got: {b}"


def test_destination1():
    lat1, lon1 = math.radians(53.32055556), math.radians(1.72972222)
    bearing = math.radians(96.02166667)
    dist = 124800
    lat2, lon2 = dll.destination_radians(lat1, lon1, bearing, dist)
    lat2, lon2 = (math.degrees(lat2), math.degrees(lon2))
    print(lat2, lon2)
    assert (lat2, lon2) == (53.188269553709034, 3.592721390871882), f"Got ({lat2}, {lon2})"


def test_distance_segment_to_segment1():
    f1 = (50.900393, 4.728607)
    f2 = (50.900389, 4.734047)
    t1 = (50.898538, 4.726107)
    t2 = (50.898176, 4.735463)
    d, pf, pt, u_f, u_t = dll.distance_segment_to_segment(f1, f2, t1, t2)
    if directory:
        plot_distance_segment_to_segment_latlon(f1, f2, t1, t2, pf, pt, "test_distance_segment_to_segment1")
    assert d == pytest.approx(216.60187728486514)
    assert pf == pytest.approx((50.900392999999994, 4.728607))
    assert pt == pytest.approx((50.898448650708666, 4.728418070396815))
    assert u_f == pytest.approx(0)
    assert u_t == pytest.approx(0.2470133466162735)


def test_distance_segment_to_segment2():
    f1 = (0, 0)
    f2 = (-0.43072496752146333, 381.4928613075559)
    t1 = (-206.26362055248765, -175.32538004745732)
    t2 = (-246.4746107556939, 480.8174213050763)
    d, pf, pt, u_f, u_t = de.distance_segment_to_segment(f1, f2, t1, t2)
    if directory:
        plot_distance_segment_to_segment_euc(f1, f2, t1, t2, pf, pt, "test_distance_segment_to_segment2")
    assert d == pytest.approx(216.60187728486514)
    assert pf == pytest.approx((0.0, 0.0))
    assert pt == pytest.approx((-216.1962718133358, -13.249350827191222))
    assert u_f == pytest.approx(0)
    assert u_t == pytest.approx(0.2470133466162735)


def test_distance_segment_to_segment3():
    f1 = (50.87205, 4.66089)
    f2 = (50.874550000000006, 4.672980000000001)
    t1 = (50.8740376, 4.6705204)
    t2 = (50.8741866999999, 4.67119980000001)
    d, pf, pt, u_f, u_t = dll.distance_segment_to_segment(f1, f2, t1, t2)
    if directory:
        plot_distance_segment_to_segment_latlon(f1, f2, t1, t2, pf, pt, "test_distance_segment_to_segment3")
    assert d == pytest.approx(0)
    assert pf == pytest.approx((50.87410572908839, 4.670830969750696))
    assert pt == pytest.approx((50.87410575464133, 4.670830955670548))
    assert u_f == pytest.approx(0.8222551304652699)
    assert u_t == pytest.approx(0.4571036354431931)


def test_distance_segment_to_segment4():
    f1 = (0, 0)
    f2 = (278.05674689789083, 848.3102386968303)
    t1 = (221.055090540802, 675.7367042826397)
    t2 = (237.6344733521503, 723.4080418578025)
    d, pf, pt, u_f, u_t = de.distance_segment_to_segment(f1, f2, t1, t2)
    if directory:
        plot_distance_segment_to_segment_euc(f1, f2, t1, t2, pf, pt, "test_distance_segment_to_segment4")
    assert d == pytest.approx(0)
    assert pf == pytest.approx((228.63358669727376, 697.5274459946864))
    assert pt == pytest.approx((228.63358669727376, 697.5274459946864))
    assert u_f == pytest.approx(0.8222551304652699)
    assert u_t == pytest.approx(0.4571036354431931)


def test_distance_point_to_segment1():
    locs = [
        (47.6373, -122.0950167),
        (47.6369, -122.0950167),
        (47.6369, -122.0959167),
        (47.6369, -122.09422),
        (47.6369, -122.09400),
        (47.6375, -122.09505)
    ]
    loc_a = (47.6372498273849, -122.094900012016)
    loc_b = (47.6368394494057, -122.094280421734)
    segments = []
    for lat_a, lat_b in itertools.product((loc_a[0], loc_b[0]), repeat=2):
        for lon_a, lon_b in itertools.product((loc_a[1], loc_b[1]), repeat=2):
            segments.append(((lat_a, lon_a), (lat_b, lon_b)))
    # segments = [(loc_a, loc_b)]

    for constrain in [True, False]:
        for loc_idx, loc in enumerate(locs):
            for seg_idx, (loc_a, loc_b) in enumerate(segments):
                dist1, pi1, ti1 = dll.distance_point_to_segment(loc, loc_a, loc_b, constrain=constrain)
                dist2, pi2, ti2 = dll.distance_point_to_segment(loc, loc_b, loc_a, constrain=constrain)
                if directory:
                    plot_distance_point_to_segment_latlon(loc, loc_a, loc_b, pi1,
                                                          f"point_to_segment_{loc_idx}_{seg_idx}_{constrain}.png")
                assert dist1 == pytest.approx(dist2), \
                    f"Locs[{loc_idx},{seg_idx},{constrain}]: Distances different, {dist1} != {dist2}"
                assert pi1[0] == pytest.approx(pi2[0]), \
                    f"Locs[{loc_idx},{seg_idx},{constrain}]: y coord different, {pi1[0]} != {pi2[0]}"
                assert pi1[1] == pytest.approx(pi2[1]), \
                    f"Locs[{loc_idx},{seg_idx},{constrain}]: y coord different, {pi1[1]} != {pi2[1]}"


def plot_distance_point_to_segment_latlon(f, t1, t2, pt, fn):
    import smopy
    import matplotlib.pyplot as plt
    lat_min = min(f[0], t1[0], t2[0])
    lat_max = max(f[0], t1[0], t2[0])
    lon_min = min(f[1], t1[1], t2[1])
    lon_max = max(f[1], t1[1], t2[1])
    bb = [lat_min, lon_min, lat_max, lon_max]
    m = smopy.Map(bb)
    ax = m.show_mpl(figsize=(10, 10))
    p1 = m.to_pixels(t1)
    p2 = m.to_pixels(t2)
    p3 = m.to_pixels(f)
    p4 = m.to_pixels(pt)
    ax.plot([p1[0], p2[0]], [p1[1], p2[1]], 'o-', color="black")
    ax.plot([p3[0]], [p3[1]], 'o-', color="black")
    ax.plot([p3[0], p4[0]], [p3[1], p4[1]], '--', color="red")
    plt.savefig(str(directory / fn))
    plt.close(plt.gcf())


def plot_distance_segment_to_segment_latlon(f1, f2, t1, t2, pf, pt, fn):
    import smopy
    import matplotlib.pyplot as plt
    lat_min = min(f1[0], f2[0], t1[0], t2[0])
    lat_max = max(f1[0], f2[0], t1[0], t2[0])
    lon_min = min(f1[1], f2[1], t1[1], t2[1])
    lon_max = max(f1[1], f2[1], t1[1], t2[1])
    bb = [lat_min, lon_min, lat_max, lon_max]
    m = smopy.Map(bb)
    ax = m.show_mpl(figsize=(10, 10))
    p1 = m.to_pixels(f1)
    p2 = m.to_pixels(f2)
    p3 = m.to_pixels(t1)
    p4 = m.to_pixels(t2)
    p5 = m.to_pixels(pf)
    p6 = m.to_pixels(pt)
    ax.plot([p1[0], p2[0]], [p1[1], p2[1]], 'o-')
    ax.plot([p3[0], p4[0]], [p3[1], p4[1]], 'o-')
    ax.plot([p5[0], p6[0]], [p5[1], p6[1]], 'x-')
    plt.savefig(str(directory / fn))
    plt.close(plt.gcf())


def plot_distance_segment_to_segment_euc(f1, f2, t1, t2, pf, pt, fn):
    import matplotlib.pyplot as plt
    fig, ax = plt.subplots(1, 1, figsize=(10, 10))
    ax.plot([f1[1], f2[1]], [f1[0], f2[0]], 'o-')
    ax.plot([t1[1], t2[1]], [t1[0], t2[0]], 'o-')
    ax.plot([pf[1], pt[1]], [pf[0], pt[0]], 'x-')
    ax.axis('equal')
    ax.set_aspect('equal')
    plt.savefig(str(directory / fn))
    plt.close(plt.gcf())


if __name__ == "__main__":
    # mm.matching.logger.setLevel(logging.INFO)
    mm.logger.setLevel(logging.DEBUG)
    mm.logger.addHandler(logging.StreamHandler(sys.stdout))
    directory = Path(os.environ.get('TESTDIR', Path(__file__).parent))
    print(f"Saving files to {directory}")
    # test_path_to_gpx()
    test_grs80()
    # test_distance1()
    # test_bearing1()
    # test_destination1()
    # test_distance_segment_to_segment1()
    # test_distance_segment_to_segment2()
    # test_distance_segment_to_segment3()
    # test_distance_segment_to_segment4()
    # test_distance_point_to_segment1()



# encoding: utf-8
"""
leuvenmapmatching.util.gpx
~~~~~~~~~~~~~~~~~~~~~~~~~~

Some additional functions to interact with the gpx library.

:author: Wannes Meert
:copyright: Copyright 2015-2018 DTAI, KU Leuven and Sirris.
:license: Apache License, Version 2.0, see LICENSE for details.
"""
import logging

import gpxpy
import gpxpy.gpx


logger = logging.getLogger("be.kuleuven.cs.dtai.mapmatching")


def gpx_to_path(gpx_file):
    gpx_fh = open(gpx_file)
    track = None
    try:
        gpx = gpxpy.parse(gpx_fh)
        if len(gpx.tracks) == 0:
            logger.error('No tracks found in GPX file (<trk> tag missing?): {}'.format(gpx_file))
            return None
        logger.info("Read gpx file: {} points, {} tracks, {} segments".format(
            gpx.get_points_no(), len(gpx.tracks), len(gpx.tracks[0].segments)))
        track = [(p.latitude, p.longitude, p.time) for p in gpx.tracks[0].segments[0].points]
    finally:
        gpx_fh.close()
    return track


def path_to_gpx(path, filename=None):
    gpx = gpxpy.gpx.GPX()

    # Create first track in our GPX:
    gpx_track = gpxpy.gpx.GPXTrack()
    gpx.tracks.append(gpx_track)

    # Create first segment in our GPX track:
    gpx_segment = gpxpy.gpx.GPXTrackSegment()
    gpx_track.segments.append(gpx_segment)

    gpx_segment.points = [(gpxpy.gpx.GPXTrackPoint(lat, lon, time=time)) for (lat, lon, time) in path]

    if filename:
        with open(filename, 'w') as gpx_fh:
            gpx_fh.write(gpx.to_xml())

    return gpx



# encoding: utf-8
"""
leuvenmapmatching.util.kalman
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

:author: Wannes Meert
:copyright: Copyright 2015-2018 DTAI, KU Leuven and Sirris.
:license: Apache License, Version 2.0, see LICENSE for details.
"""
import logging

from pykalman import KalmanFilter
import numpy as np


logger = logging.getLogger("be.kuleuven.cs.dtai.mapmatching")


def smooth_path(path, dt=1, obs_noise=1e-4, loc_var=1e-4, vel_var=1e-6, kf=None,
                rm_outliers=False, use_euclidean=True, n_iter=1000):
    """Apply Kalman filtering. Assumes data with a constant sample rate.

    Inspired by https://github.com/FlorianWilhelm/gps_data_with_python

    :param path:
    :param dt: Sample interval in seconds
    :param obs_noise: Observation noise (default=1e-4, approx 10-30m)
    :param loc_var: estimated location variance
    :param vel_var: estimated velocity variance
    :param kf: Trained Kalman filter
    :param rm_outliers: Remove outliers based on Kalman prediction
        True or 1 will be removal, 2 will also retrain after removal
    :param use_euclidean:
    :param n_iter: Kalman iterations
    :return:
    """
    path = np.array(path)
    if kf is None:
        # state is (x, y, v_x, v_y)
        F = np.array([[1, 0, dt, 0],
                      [0, 1, 0,  dt],
                      [0, 0, 1,  0],
                      [0, 0, 0,  1]])

        # observations is (x, y)
        H = np.array([[1, 0, 0, 0],
                      [0, 1, 0, 0]])

        R = np.diag([obs_noise, obs_noise]) ** 2

        initial_state_mean = np.hstack([path[0, :2], 2 * [0.]])
        initial_state_covariance = np.diag([loc_var, loc_var, vel_var, vel_var]) ** 2

        kf = KalmanFilter(transition_matrices=F,
                          observation_matrices=H,
                          observation_covariance=R,
                          initial_state_mean=initial_state_mean,
                          initial_state_covariance=initial_state_covariance,
                          em_vars=['transition_covariance'])

        if n_iter > 0:
            logger.debug("Start learning")
            kf = kf.em(path[:, :2], n_iter=n_iter)

    state_means, state_vars = kf.smooth(path[:, :2])

    if use_euclidean:
        from .dist_euclidean import distance
        distance_f = distance
    else:
        from .dist_latlon import distance
        distance_f = distance
    if rm_outliers:
        path_ma = np.ma.asarray(path[:, :2])
        for idx in range(path.shape[0]):
            d = distance_f(path[idx, :2], state_means[idx, :2])
            if d > obs_noise * 2:
                logger.debug("Rm point {}".format(idx))
                path_ma[idx] = np.ma.masked
        if rm_outliers == 2:
            logger.debug("Retrain")
            kf = kf.em(path_ma, n_iter=n_iter)
        state_means, state_vars = kf.smooth(path_ma)

    return state_means, state_vars, kf



#!/usr/bin/env python3
# encoding: utf-8
"""
tests.test_nonemitting_circle
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

:author: Wannes Meert
:copyright: Copyright 2018 DTAI, KU Leuven and Sirris.
:license: Apache License, Version 2.0, see LICENSE for details.
"""
import sys, os
import logging
import math
from pathlib import Path
import numpy as np

try:
    import leuvenmapmatching as mm
except ImportError:
    sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), os.pardir)))
    import leuvenmapmatching as mm
from leuvenmapmatching.matcher.simple import SimpleMatcher
from leuvenmapmatching.matcher.distance import DistanceMatcher
from leuvenmapmatching.map.inmem import InMemMap


directory = None


def setup_map(disconnect=True):
    theta = np.linspace(0, 2 * math.pi, 4 * 5 + 1)[:-1]

    ox = 0.1 + np.cos(theta * 0.95)
    oy = np.sin(theta * 1)
    path1 = list(zip(ox, oy))  # all observations
    path2 = [(x, y) for x, y in zip(ox, oy) if x > -0.60]

    nx = np.cos(theta)
    ny = np.sin(theta)
    nl = [f"N{i}" for i in range(len(nx))]
    graph = {}
    for i, (x, y, l) in enumerate(zip(nx, ny, nl)):
        if disconnect:
            edges = []
            if i != len(nx) - 1:
                edges.append(nl[(i + 1) % len(nl)])
            if i != 0:
                edges.append(nl[(i - 1) % len(nl)])
        else:
            edges = [nl[(i - 1) % len(nl)], nl[(i + 1) % len(nl)]]
        graph[l] = ((x, y), edges)
    graph["M"] = ((0, 0), ["N5", "N15"])
    graph["N5"][1].append("M")
    graph["N15"][1].append("M")
    print(graph)

    path_sol = nl
    if not disconnect:
        path_sol += ["N0"]

    mapdb = InMemMap("map", graph=graph, use_latlon=False)
    return mapdb, path1, path2, path_sol


def visualize_map():
    if directory is None:
        return
    import matplotlib.pyplot as plt
    import leuvenmapmatching.visualization as mm_vis

    mapdb, path1, path2, path_sol = setup_map()

    fig, ax = plt.subplots(ncols=1, nrows=1, figsize=(10, 5))
    mm_vis.plot_map(mapdb, path=path1, ax=ax, show_labels=True)
    fig.savefig(str(directory / 'test_nonemitting_circle_map_path1.png'))
    plt.close(fig)

    fig, ax = plt.subplots(ncols=1, nrows=1, figsize=(10, 10))
    mm_vis.plot_map(mapdb, path=path2, ax=ax, show_labels=True)
    fig.savefig(str(directory / 'test_nonemitting_circle_map_path2.png'))
    plt.close(fig)


def visualize_path(matcher, mapdb, name="test"):
    import matplotlib.pyplot as plt
    from leuvenmapmatching import visualization as mmviz
    fig, ax = plt.subplots(1, 1, figsize=(10, 10))
    mmviz.plot_map(mapdb, matcher=matcher, ax=ax,
                   show_labels=True, show_matching=True, show_graph=True,
                   linewidth=2)
    fn = directory / f"test_nonemitting_circle_{name}_map.png"
    fig.savefig(str(fn))
    plt.close(fig)
    print(f"saved to {fn}")


def test_path1():
    mapdb, path1, path2, path_sol = setup_map()
    matcher = SimpleMatcher(mapdb, max_dist_init=1, min_prob_norm=0.5, obs_noise=0.5,
                            non_emitting_states=True)
    matcher.match(path1, unique=True)
    path_pred = matcher.path_pred_onlynodes
    if directory:
        visualize_path(matcher, mapdb, name="testpath1")
    assert path_pred == path_sol, f"Nodes not equal:\n{path_pred}\n{path_sol}"


def test_path1_dist():
    mapdb, path1, path2, path_sol = setup_map()
    matcher = DistanceMatcher(mapdb, max_dist_init=1, min_prob_norm=0.8, obs_noise=0.5,
                              non_emitting_states=True)
    matcher.match(path1, unique=True)
    path_pred = matcher.path_pred_onlynodes
    if directory:
        visualize_path(matcher, mapdb, name="test_path1_dist")
    assert path_pred == path_sol, f"Nodes not equal:\n{path_pred}\n{path_sol}"


def test_path2():
    mapdb, path1, path2, _ = setup_map()
    path_sol = [f"N{i}" for i in range(20)]
    matcher = SimpleMatcher(mapdb, max_dist_init=0.2, min_prob_norm=0.1,
                                  obs_noise=0.1, obs_noise_ne=1,
                                  non_emitting_states=True, only_edges=True)
    path_pred = matcher.match(path2, unique=True)
    print(path_pred)
    path_pred = matcher.path_pred_onlynodes
    if directory:
        matcher.print_lattice_stats(verbose=True)
        matcher.print_lattice()
        print("Best path through lattice:")
        for m in matcher.lattice_best:
            print(m)
        visualize_path(matcher, mapdb, name="testpath2")
    assert path_pred == path_sol, f"Nodes not equal:\n{path_pred}\n{path_sol}"


def test_path2_dist():
    mapdb, path1, path2, _ = setup_map()
    path_sol = [f"N{i}" for i in range(20)]
    matcher = DistanceMatcher(mapdb, max_dist_init=0.2, min_prob_norm=0.1,
                              obs_noise=0.1, obs_noise_ne=1,
                              non_emitting_states=True)
    matcher.match(path2)
    path_pred = matcher.path_pred_onlynodes
    if directory:
        visualize_path(matcher, mapdb, name="test_path2_dist")
    assert path_pred == path_sol, f"Nodes not equal:\n{path_pred}\n{path_sol}"


if __name__ == "__main__":
    # mm.matching.logger.setLevel(logging.INFO)
    mm.logger.setLevel(logging.DEBUG)
    mm.logger.addHandler(logging.StreamHandler(sys.stdout))
    directory = Path(os.environ.get('TESTDIR', Path(__file__).parent))
    print(f"Saving files to {directory}")
    # visualize_map()
    test_path1()
    # test_path1_dist()
    # test_path2()
    # test_path2_dist()



#!/usr/bin/env python3
# encoding: utf-8
"""
tests.test_path_newsonkrumm2009
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Based on the data available at:
https://www.microsoft.com/en-us/research/publication/hidden-markov-map-matching-noise-sparseness/

Notes:

* There is a 'bug' in the map available from the website.
  Multiple segments (streets) in the map are not connected but have overlapping, but
  disconnected, nodes.
  For example, the following nodes are on the same location and
  should be connected because the given path runs over this road:
  - 884147801204 and 884148400033
  - 884148100260 and 884148001002
* The path is missing a number of observations. For those parts non-emitting nodes are required.
  This occurs at:
  - 2770:2800 (index 2659 is start)
  - 2910:2929

:author: Wannes Meert
:copyright: Copyright 2018 DTAI, KU Leuven and Sirris.
:license: Apache License, Version 2.0, see LICENSE for details.
"""
import os
import sys
import logging
import pickle
from pathlib import Path
import csv
from datetime import datetime
from itertools import product
import pytest
import leuvenmapmatching as mm
from leuvenmapmatching.matcher import base
from leuvenmapmatching.matcher.distance import DistanceMatcher
from leuvenmapmatching.map.sqlite import SqliteMap
import leuvenmapmatching.visualization as mm_viz
MYPY = False
if MYPY:
    from typing import List, Tuple


logger = mm.logger
this_path = Path(os.path.realpath(__file__)).parent / "rsrc" / "newson_krumm_2009"
gps_data = this_path / "gps_data.txt"
gps_data_pkl = gps_data.with_suffix(".pkl")
ground_truth_route = this_path / "ground_truth_route.txt"
road_network = this_path / "road_network.txt"
road_network_zip = this_path / "road_network.zip"
road_network_db = road_network.with_suffix(".sqlite")

directory = None
base.default_label_width = 34


def read_gps(route_fn):
    route = []
    with route_fn.open("r") as route_f:
        reader = csv.reader(route_f, delimiter='\t')
        next(reader)
        for row in reader:
            date, time, lat, lon = row[:4]
            date_str = date + " " + time
            ts = datetime.strptime(date_str, '%d-%b-%Y %H:%M:%S')
            lat = float(lat)
            lon = float(lon)
            route.append((lat, lon, ts))
    logger.debug(f"Read GPS trace of {len(route)} points")
    return route


def read_paths(paths_fn):
    paths = []
    with paths_fn.open("r") as paths_f:
        reader = csv.reader(paths_f, delimiter='\t')
        next(reader)
        for row in reader:
            pathid, trav = row[:2]
            pathid = int(pathid)
            trav = int(trav)
            paths.append((pathid, trav))
    logger.debug(f"Read correct trace of {len(paths)} nodes")
    return paths


def parse_linestring(line):
    # type: (str) -> List[Tuple[float, float]]
    line = line[line.index("(") + 1:line.index(")")]
    latlons = []
    for lonlat in line.split(", "):
        lon, lat = lonlat.split(" ")
        latlons.append((float(lat), float(lon)))
    return latlons


def read_map(map_fn):
    logger.debug(f"Reading map ...")
    mmap = SqliteMap("road_network", use_latlon=True, dir=this_path)
    node_cnt = 0
    edge_cnt = 0
    # new_node_id = 1000000000000
    new_node_id = 1
    with map_fn.open("r") as map_f:
        reader = csv.reader(map_f, delimiter='\t')
        next(reader)
        for row in reader:
            eid, nf, nt, twoway, speed, length, innernodes = row
            eid = int(eid)
            nf = int(nf)
            nt = int(nt)
            length = int(length)
            twoway = int(twoway)
            speed = float(speed)
            if twoway == 0:
                twoway = False
            elif twoway == 1:
                twoway = True
            else:
                raise Exception(f"Unknown value for twoway: {twoway}")
            innernodes = parse_linestring(innernodes)
            # Add nodes to map
            mmap.add_node(nf, innernodes[0], ignore_doubles=True, no_index=True, no_commit=True)
            mmap.add_node(nt, innernodes[-1], ignore_doubles=True, no_index=True, no_commit=True)
            node_cnt += 2
            prev_node = nf
            assert(length < 1000)
            idx = 1
            for innernode in innernodes[1:-1]:
                # innernode_id = nf * 1000 + idx
                innernode_id = new_node_id
                new_node_id += 1
                mmap.add_node(innernode_id, innernode, no_index=True, no_commit=True)  # Should not be double
                node_cnt += 1
                mmap.add_edge(prev_node, innernode_id, speed=speed, edge_type=0,
                              path=eid, pathnum=idx, no_index=True, no_commit=True)
                edge_cnt += 1
                if twoway:
                    mmap.add_edge(innernode_id, prev_node, speed=speed, edge_type=0,
                                  path=eid, pathnum=-idx, no_index=True, no_commit=True)
                    edge_cnt += 1
                prev_node = innernode_id
                idx += 1
            mmap.add_edge(prev_node, nt, speed=speed, edge_type=0,
                          path=eid, pathnum=idx, no_index=True, no_commit=True)
            edge_cnt += 1
            if twoway:
                mmap.add_edge(nt, prev_node, speed=speed, edge_type=0,
                              path=eid, pathnum=-idx, no_index=True, no_commit=True)
                edge_cnt += 1
            if node_cnt % 100000 == 0:
                mmap.db.commit()
    logger.debug(f"... done: {node_cnt} nodes and {edge_cnt} edges")
    mmap.reindex_nodes()
    mmap.reindex_edges()
    assert(new_node_id < 100000000000)
    return mmap


def correct_map(mmap):
    """Add edges between nodes with degree > 2 that are on the exact same location.
    This ignore that with bridges, the roads might not be connected. But we need a correct
    because the dataset has a number of interrupted paths.
    """
    def correct_edge(labels):
        labels = [label for label in labels if label > 100000000000]
        logger.info(f"Add connections between {labels}")
        for l1, l2 in product(labels, repeat=2):
            mmap.add_edge(l1, l2, edge_type=1)
    mmap.find_duplicates(func=correct_edge)


def load_data():
    max_route_length = None  # 200

    # Paths
    if not ground_truth_route.exists():
        import requests
        url = f'https://www.microsoft.com/en-us/research/uploads/prod/2017/07/ground_truth_route.txt'
        logger.debug("Download gound_truth_route.txt from microsoft.com")
        r = requests.get(url, stream=True)
        with ground_truth_route.open('wb') as ofile:
            for chunk in r.iter_content(chunk_size=1024):
                if chunk:
                    ofile.write(chunk)
    paths = read_paths(ground_truth_route)

    # Map
    if road_network_db.exists():
        map_con = SqliteMap.from_file(road_network_db)
        logger.debug(f"Read road network from db file {road_network_db} ({map_con.size()} nodes)")
    else:
        if not road_network.exists():
            import requests
            url = f'https://www.microsoft.com/en-us/research/uploads/prod/2017/07/road_network.zip'
            logger.debug("Download road_network.zip from microsoft.com")
            r = requests.get(url, stream=True)
            with road_network_zip.open('wb') as ofile:
                for chunk in r.iter_content(chunk_size=1024):
                    if chunk:
                        ofile.write(chunk)
            import zipfile
            logger.debug("Unzipping road_network.zip")
            with zipfile.ZipFile(str(road_network_zip), "r") as zip_ref:
                zip_ref.extractall(str(road_network_zip.parent))
        map_con = read_map(road_network)
        correct_map(map_con)
        logger.debug(f"Create road network to db file {map_con.db_fn} ({map_con.size()} nodes)")

    # Route
    if gps_data_pkl.exists():
        with gps_data_pkl.open("rb") as ifile:
            route = pickle.load(ifile)
        logger.debug(f"Read gps route from file ({len(route)} points)")
    else:
        if not gps_data.exists():
            import requests
            url = f'https://www.microsoft.com/en-us/research/uploads/prod/2017/07/gps_data.txt'
            logger.debug("Download gps_data.txt from microsoft.com")
            r = requests.get(url, stream=True)
            with gps_data.open('wb') as ofile:
                for chunk in r.iter_content(chunk_size=1024):
                    if chunk:
                        ofile.write(chunk)
        route = read_gps(gps_data)
        if max_route_length:
            route = route[:max_route_length]
        with gps_data_pkl.open("wb") as ofile:
            pickle.dump(route, ofile)
    route = [(lat, lon) for lat, lon, _ in route]

    return paths, map_con, route


def test_route_slice1():
    if directory:
        import matplotlib.pyplot as plt
    nodes, map_con, route = load_data()
    zoom_path = True

    matcher = DistanceMatcher(map_con, min_prob_norm=0.001,
                              max_dist=200,
                              dist_noise=6, dist_noise_ne=12,
                              obs_noise=30, obs_noise_ne=150,
                              non_emitting_states=True)
    route_slice = route[2657:2662]
    matcher.match(route_slice)
    path_pred = matcher.path_pred_onlynodes
    path_sol = [172815, 172816, 172817, 172818, 172819, 172820, 172821, 172822, 172823, 172824,
                172825, 172826, 172827, 172828, 172829, 172830, 884148100261, 172835, 172836,
                172837, 884148100254, 172806, 884148100255, 172807]  # Can change when building db
    assert len(path_pred) == len(path_sol)


def test_bug1():
    map_con = SqliteMap("map", use_latlon=True)
    map_con.add_nodes([
        (1, (47.590439915657, -122.238368690014)),
        (2, (47.5910192728043, -122.239519357681)),
        (3, (47.5913706421852, -122.240168452263))
    ])
    map_con.add_edges([
        (1, 2),
        (2, 3)
    ])
    path = [
        # (47.59043333, -122.2384167),
        (47.59058333, -122.2387),
        (47.59071667, -122.2389833),
        (47.59086667, -122.2392667),
        (47.59101667, -122.23955),
        (47.59115,    -122.2398333)
    ]
    path_sol = [(1, 2), (2, 3)]
    matcher = DistanceMatcher(map_con, min_prob_norm=0.001,
                              max_dist=200, obs_noise=4.07,
                              non_emitting_states=True)
    matcher.match(path, unique=True)
    path_pred = matcher.path_pred
    if directory:
        import matplotlib.pyplot as plt
        matcher.print_lattice_stats()
        logger.debug("Plotting post map ...")
        fig = plt.figure(figsize=(100, 100))
        ax = fig.get_axes()
        mm_viz.plot_map(map_con, matcher=matcher, use_osm=True, ax=ax,
                        show_lattice=False, show_labels=True, show_graph=True, zoom_path=True,
                        show_matching=True)
        plt.savefig(str(directory / "test_newson_bug1.png"))
        plt.close(fig)
        logger.debug("... done")
    assert path_pred == path_sol, f"Edges not equal:\n{path_pred}\n{path_sol}"


@pytest.mark.skip(reason="Takes a long time")
def test_route():
    if directory:
        import matplotlib.pyplot as plt
    else:
        plt = None
    paths, map_con, route = load_data()
    route = [(lat, lon) for lat, lon, _ in route]
    zoom_path = True
    # zoom_path = slice(2645, 2665)
    slice_route = None
    # slice_route = slice(650, 750)
    # slice_route = slice(2657, 2662)  # First location where some observations are missing
    # slice_route = slice(2770, 2800)  # Observations are missing
    # slice_route = slice(2910, 2950)  # Interesting point
    # slice_route = slice(2910, 2929)  # Interesting point
    # slice_route = slice(6825, 6833)  # Outlier observation

    # if directory is not None:
    #     logger.debug("Plotting pre map ...")
    #     mm_viz.plot_map(map_con_latlon, path=route_latlon, use_osm=True,
    #                     show_lattice=False, show_labels=False, show_graph=False, zoom_path=zoom_path,
    #                     filename=str(directory / "test_newson_route.png"))
    #     logger.debug("... done")

    matcher = DistanceMatcher(map_con, min_prob_norm=0.0001,
                              max_dist=200,
                              dist_noise=15, dist_noise_ne=30,
                              obs_noise=30, obs_noise_ne=150,
                              non_emitting_states=True)

    if slice_route is None:
        pkl_fn = this_path / "nodes_pred.pkl"
        if pkl_fn.exists():
            with pkl_fn.open("rb") as pkl_file:
                logger.debug(f"Reading predicted nodes from pkl file")
                route_nodes = pickle.load(pkl_file)
        else:
            matcher.match(route)
            route_nodes = matcher.path_pred_onlynodes
            with pkl_fn.open("wb") as pkl_file:
                pickle.dump(route_nodes, pkl_file)
        from leuvenmapmatching.util.evaluation import route_mismatch_factor
        print(route_nodes[:10])
        # route_edges = map_con.nodes_to_paths(route_nodes)
        # print(route_edges[:10])
        grnd_paths, _ = zip(*paths)
        print(grnd_paths[:10])
        route_paths = map_con.nodes_to_paths(route_nodes)
        print(route_paths[:10])

        logger.debug(f"Compute route mismatch factor")
        factor, cnt_matches, cnt_mismatches, total_length, mismatches = \
            route_mismatch_factor(map_con, route_paths, grnd_paths,window=None, keep_mismatches=True)
        logger.debug(f"factor = {factor}, "
                     f"cnt_matches = {cnt_matches}/{cnt_mismatches} of {len(grnd_paths)}/{len(route_paths)}, "
                     f"total_length = {total_length}\n"
                     f"mismatches = " + " | ".join(str(v) for v in mismatches))
    else:
        _, last_idx = matcher.match(route[slice_route])
        logger.debug(f"Last index = {last_idx}")

    # matcher.match(route[2657:2662])  # First location where some observations are missing
    # matcher.match(route[2770:2800])  # Observations are missing
    # matcher.match(route[2910:2950])  # Interesting point
    # matcher.match(route[2910:2929])  # Interesting point
    # matcher.match(route[6000:])
    path_pred = matcher.path_pred_onlynodes

    if directory:
        matcher.print_lattice_stats()
        logger.debug("Plotting post map ...")
        fig = plt.figure(figsize=(200, 200))
        ax = fig.get_axes()
        mm_viz.plot_map(map_con, matcher=matcher, use_osm=True, ax=ax,
                        show_lattice=False, show_labels=True, zoom_path=zoom_path,
                        show_matching=True, show_graph=False)
        plt.savefig(str(directory / "test_newson_route_matched.png"))
        plt.close(fig)
        logger.debug("... done")
        logger.debug("Best path:")
        for m in matcher.lattice_best:
            logger.debug(m)

    print(path_pred)


@pytest.mark.skip(reason="Takes a too long")
def test_bug2():
    from leuvenmapmatching.util.openstreetmap import locations_to_map
    map_con = SqliteMap("map", use_latlon=True, dir=directory)
    path = [
        (50.87205, 4.66089), (50.874550000000006, 4.672980000000001), (50.87538000000001, 4.67698),
        (50.875800000000005, 4.6787600000000005), (50.876520000000006, 4.6818), (50.87688000000001, 4.683280000000001),
        (50.87814, 4.68733), (50.87832, 4.68778), (50.87879, 4.68851), (50.87903000000001, 4.68895),
        (50.879560000000005, 4.689170000000001), (50.87946, 4.6900900000000005),
        (50.879290000000005, 4.6909600000000005), (50.87906, 4.6921800000000005), (50.87935, 4.6924),
        (50.879720000000006, 4.69275), (50.88002, 4.6930700000000005), (50.880430000000004, 4.693440000000001),
        (50.880660000000006, 4.69357), (50.880660000000006, 4.6936100000000005), (50.88058, 4.694640000000001),
        (50.88055000000001, 4.69491), (50.88036, 4.696160000000001), (50.88009, 4.697550000000001),
        (50.87986, 4.6982800000000005), (50.879720000000006, 4.698790000000001), (50.87948, 4.699730000000001),
        (50.87914000000001, 4.6996400000000005), (50.87894000000001, 4.6995000000000005),
        (50.878800000000005, 4.699350000000001), (50.8785, 4.6991000000000005), (50.87841, 4.6990300000000005)
    ]
    locations_to_map(path, map_con, filename=directory / "osm.xml")
    path_sol = [(5777282112, 2633552218), (2633552218, 5777282111), (5777282111, 5777282110), (5777282110, 1642021707),
                (1642021707, 71361087), (71361087, 71364203), (71364203, 1151697757), (1151697757, 1647339017),
                (1647339017, 1647339030), (1647339030, 2058510349), (2058510349, 2633552212), (2633552212, 1380538577),
                (1380538577, 1439572271), (1439572271, 836434313), (836434313, 2633771041), (2633771041, 5042874484),
                (5042874484, 5042874485), (5042874485, 2518922583), (2518922583, 2659762546), (2659762546, 5777282063),
                (5777282063, 2633771037), (2633771037, 2633771035), (2633771035, 2633771033), (2633771033, 1151668705),
                (1151668705, 2633771094), (2633771094, 1151668722), (1151668722, 1151668724), (1151668724, 5543948222),
                (5543948222, 2058481517), (2058481517, 16933576), (16933576, 5543948221), (5543948221, 2518923620),
                (2518923620, 5543948020), (5543948020, 5543948019), (5543948019, 18635886), (18635886, 18635887),
                (18635887, 1036909153), (1036909153, 2658942230), (2658942230, 1001099975), (1001099975, 16933574),
                (16933574, 1125604152), (1125604152, 5543948238), (5543948238, 1125604150), (1125604150, 1125604148),
                (1125604148, 2634195334), (2634195334, 2087854243), (2087854243, 5543948237), (5543948237, 160226603),
                (160226603, 180130266), (180130266, 5543948227), (5543948227, 5543948226), (5543948226, 1195681902),
                (1195681902, 101135392), (101135392, 2606704673), (2606704673, 18635977), (18635977, 1026111708),
                (1026111708, 1026111631), (1026111631, 16571375), (16571375, 2000680621), (2000680621, 999580042),
                (999580042, 16571370), (16571370, 2000680620), (2000680620, 5078692402), (5078692402, 5543948008),
                (5543948008, 16571371), (16571371, 999579936), (999579936, 2639836143), (2639836143, 5543948014),
                (5543948014, 5222992316), (5222992316, 30251323), (30251323, 159701080), (159701080, 3173217124),
                (3173217124, 1165209673), (1165209673, 1380538689), (1380538689, 2878334668), (2878334668, 2871137399),
                (2871137399, 2876902981), (2876902981, 2873624508), (2873624508, 2873624509), (2873624509, 2899666507),
                (2899666507, 2899666518), (2899666518, 2899666513), (2899666513, 2903073945), (2903073945, 2903073951),
                (2903073951, 1380538681), (1380538681, 2914810627), (2914810627, 2914810618), (2914810618, 2914810607),
                (2914810607, 2914810604), (2914810604, 2914810483), (2914810483, 2914810462), (2914810462, 2914810464),
                (2914810464, 1312433523), (1312433523, 20918594), (20918594, 2634267817), (2634267817, 2967425445),
                (2967425445, 3201523879), (3201523879, 157217466), (157217466, 2963305939), (2963305939, 3201523877),
                (3201523877, 3889275909), (3889275909, 3889275897), (3889275897, 157255077), (157255077, 30251882),
                (30251882, 157245624), (157245624, 1150903673), (1150903673, 4504936404)]
    matcher = DistanceMatcher(map_con, min_prob_norm=0.001,
                              max_dist=200, obs_noise=4.07,
                              non_emitting_states=True)
    nodes, idx = matcher.match(path, unique=True)
    path_pred = matcher.path_pred
    if directory:
        import matplotlib.pyplot as plt
        matcher.print_lattice_stats()
        logger.debug("Plotting post map ...")
        fig = plt.figure(figsize=(100, 100))
        ax = fig.get_axes()
        mm_viz.plot_map(map_con, matcher=matcher, use_osm=True, ax=ax,
                        show_lattice=False, show_labels=True, show_graph=False, zoom_path=True,
                        show_matching=True)
        plt.savefig(str(directory / "test_newson_bug1.png"))
        plt.close(fig)
        logger.debug("... done")
    assert path_pred == path_sol, f"Edges not equal:\n{path_pred}\n{path_sol}"


if __name__ == "__main__":
    logger.setLevel(logging.DEBUG)
    logger.addHandler(logging.StreamHandler(sys.stdout))
    directory = Path(os.environ.get('TESTDIR', Path(__file__).parent))
    print(f"Saving files to {directory}")
    # test_route()
    test_route_slice1()
    # test_bug1()
    # test_bug2()



#!/usr/bin/env python3
# encoding: utf-8
"""
tests.test_nonemitting
~~~~~~~~~~~~~~~~~~~~~~

:author: Wannes Meert
:copyright: Copyright 2017-2018 DTAI, KU Leuven and Sirris.
:license: Apache License, Version 2.0, see LICENSE for details.
"""
import sys
import os
import logging
from pathlib import Path

try:
    import leuvenmapmatching as mm
except ImportError:
    sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), os.pardir)))
    import leuvenmapmatching as mm
from leuvenmapmatching.matcher.distance import DistanceMatcher, DistanceMatching
from leuvenmapmatching.matcher.simple import SimpleMatcher
from leuvenmapmatching.map.inmem import InMemMap

MYPY = False
if MYPY:
    from typing import Tuple


logger = mm.logger
directory = None


def setup_map():
    path1 = [(1.8, 0.1), (1.8, 3.5), (3.0, 4.9)]  # More nodes than observations
    path2 = [(1.8, 0.1), (1.8, 2.0), (1.8, 3.5), (3.0, 4.9)]
    path_sol = ['X', 'C', 'D', 'F']
    mapdb = InMemMap("map", graph={
        "A": ((1, 1), ["B", "C", "X"]),
        "B": ((1, 3), ["A", "C", "D", "K"]),
        "C": ((2, 2), ["A", "B", "D", "E", "X", "Y"]),
        "D": ((2, 4), ["B", "C", "E", "K", "L", "F"]),
        "E": ((3, 3), ["C", "D", "F", "Y"]),
        "F": ((3, 5), ["D", "E", "L"]),
        "X": ((2, 0), ["A", "C", "Y"]),
        "Y": ((3, 1), ["X", "C", "E"]),
        "K": ((1, 5), ["B", "D", "L"]),
        "L": ((2, 6), ["K", "D", "F"])
    }, use_latlon=False)
    return mapdb, path1, path2, path_sol


def visualize_map(pathnb=1):
    mapdb, path1, path2, path_sol = setup_map()
    import leuvenmapmatching.visualization as mm_vis
    if pathnb == 2:
        path = path2
    else:
        path = path1
    mm_vis.plot_map(mapdb, path=path, show_labels=True,
                    filename=(directory / "test_nonemitting_map.png"))


def test_path1():
    mapdb, path1, path2, path_sol = setup_map()

    matcher = SimpleMatcher(mapdb, max_dist_init=1,
                                  min_prob_norm=0.5,
                                  obs_noise=0.5,
                                  non_emitting_states=True, only_edges=False)
    matcher.match(path1, unique=True)
    path_pred = matcher.path_pred_onlynodes
    if directory:
        from leuvenmapmatching import visualization as mmviz
        matcher.print_lattice_stats()
        matcher.print_lattice()
        with (directory / 'lattice_path1.gv').open('w') as ofile:
            matcher.lattice_dot(file=ofile)
        mmviz.plot_map(mapdb, matcher=matcher, show_labels=True, show_matching=True,
                       show_graph=True,
                       filename=str(directory / "test_nonemitting_test_path1.png"))
    assert path_pred == path_sol, f"Nodes not equal:\n{path_pred}\n{path_sol}"


def test_path1_inc():
    mapdb, path1, path2, path_sol = setup_map()

    matcher = SimpleMatcher(mapdb, max_dist_init=1,
                            in_prob_norm=0.5, obs_noise=0.5,
                            non_emitting_states=True, only_edges=False,
                            max_lattice_width=1)

    print('## PHASE 1 ##')
    matcher.match(path1, unique=True)
    path_pred = matcher.path_pred_onlynodes
    if directory:
        from leuvenmapmatching import visualization as mmviz
        matcher.print_lattice_stats()
        matcher.print_lattice()
        with (directory / 'lattice_path1_inc1.gv').open('w') as ofile:
            matcher.lattice_dot(file=ofile, precision=2, render=True)
        mmviz.plot_map(mapdb, matcher=matcher, show_labels=True, show_matching=True,
                       show_graph=True,
                       filename=str(directory / "test_nonemitting_test_path1_inc1.png"))

    print('## PHASE 2 ##')
    matcher.increase_max_lattice_width(3, unique=True)
    path_pred = matcher.path_pred_onlynodes
    if directory:
        from leuvenmapmatching import visualization as mmviz
        matcher.print_lattice_stats()
        matcher.print_lattice()
        with (directory / 'lattice_path1_inc2.gv').open('w') as ofile:
            matcher.lattice_dot(file=ofile, precision=2, render=True)
        mmviz.plot_map(mapdb, matcher=matcher, show_labels=True, show_matching=True,
                       show_graph=True,
                       filename=str(directory / "test_nonemitting_test_path1_inc2.png"))


    assert path_pred == path_sol, f"Nodes not equal:\n{path_pred}\n{path_sol}"


def test_path1_dist():
    mapdb, path1, path2, path_sol = setup_map()

    matcher = DistanceMatcher(mapdb, max_dist_init=1,
                              min_prob_norm=0.5,
                              obs_noise=0.5,
                              non_emitting_states=True, only_edges=True)
    matcher.match(path1, unique=True)
    path_pred = matcher.path_pred_onlynodes
    if directory:
        from leuvenmapmatching import visualization as mmviz
        matcher.print_lattice_stats()
        matcher.print_lattice()
        print("LATTICE BEST")
        for m in matcher.lattice_best:
            print(m)
        with (directory / 'lattice_path1.gv').open('w') as ofile:
            matcher.lattice_dot(file=ofile)
        mmviz.plot_map(mapdb, matcher=matcher, show_labels=True, show_matching=True, show_graph=True,
                       filename=str(directory / "test_nonemitting_test_path1_dist.png"))
    assert path_pred == path_sol, f"Nodes not equal:\n{path_pred}\n{path_sol}"


def test_path2():
    mapdb, path1, path2, path_sol = setup_map()

    matcher = SimpleMatcher(mapdb, max_dist_init=1, min_prob_norm=0.5, obs_noise=0.5,
                                  non_emitting_states=True, only_edges=False)
    matcher.match(path2, unique=True)
    path_pred = matcher.path_pred_onlynodes

    dists = matcher.path_all_distances()

    if directory:
        from leuvenmapmatching import visualization as mmviz
        matcher.print_lattice_stats()
        matcher.print_lattice()
        with (directory / 'lattice_path2.gv').open('w') as ofile:
            matcher.lattice_dot(file=ofile)
        mmviz.plot_map(mapdb, matcher=matcher, show_labels=True, show_matching=True,
                       filename=str(directory / "test_nonemitting_test_path2.png"))
    assert path_pred == path_sol, "Nodes not equal:\n{}\n{}".format(path_pred, path_sol)


def test_path2_dist():
    mapdb, path1, path2, path_sol = setup_map()

    matcher = DistanceMatcher(mapdb, max_dist_init=1, min_prob_norm=0.5,
                              obs_noise=0.5, dist_noise=0.5,
                              non_emitting_states=True)
    matcher.match(path2, unique=True)
    path_pred = matcher.path_pred_onlynodes
    if directory:
        from leuvenmapmatching import visualization as mmviz
        matcher.print_lattice_stats()
        matcher.print_lattice()
        # with (directory / 'lattice_path2.gv').open('w') as ofile:
        #     matcher.lattice_dot(file=ofile)
        mmviz.plot_map(mapdb, matcher=matcher, show_labels=True, show_matching=True,
                       filename=str(directory / "test_nonemitting_test_path2_dist.png"))
    assert path_pred == path_sol, "Nodes not equal:\n{}\n{}".format(path_pred, path_sol)


def test_path2_incremental():
    mapdb, path1, path2, path_sol = setup_map()

    matcher = SimpleMatcher(mapdb, max_dist_init=1, min_prob_norm=0.5, obs_noise=0.5,
                                  non_emitting_states=True, only_edges=False)
    matcher.match(path2[:2])
    path_pred_1 = matcher.path_pred_onlynodes
    matcher.match(path2, expand=True)
    path_pred = matcher.path_pred_onlynodes
    if directory:
        from leuvenmapmatching import visualization as mmviz
        matcher.print_lattice_stats()
        matcher.print_lattice()
        with (directory / 'lattice_path2.gv').open('w') as ofile:
            matcher.lattice_dot(file=ofile)
        mmviz.plot_map(mapdb, matcher=matcher, show_labels=True, show_matching=True,
                       filename=str(directory / "test_nonemitting_test_path2.png"))
    assert path_pred_1 == path_sol[:2], "Nodes not equal:\n{}\n{}".format(path_pred, path_sol)
    assert path_pred == path_sol, "Nodes not equal:\n{}\n{}".format(path_pred, path_sol)


def test_path_duplicate():
    from datetime import datetime
    # A path with two identical points
    path = [(0.8, 0.7), (0.9, 0.7), (1.1, 1.0), (1.2, 1.5), (1.2, 1.5), (1.2, 1.6), (1.1, 2.0),
            (1.1, 2.3), (1.3, 2.9), (1.2, 3.1), (1.5, 3.2), (1.8, 3.5), (2.0, 3.7),
            (2.1, 3.3), (2.4, 3.2), (2.6, 3.1), (2.9, 3.1), (3.0, 3.2), (3.1, 3.8),
            (3.0, 4.0), (3.1, 4.3), (3.1, 4.6), (3.0, 4.9)]

    mapdb = InMemMap("map", graph={
        "A": ((1, 1), ["B", "C"]),
        "B": ((1, 3), ["A", "C", "D"]),
        "C": ((2, 2), ["A", "B", "D", "E"]),
        "D": ((2, 4), ["B", "C", "D", "E"]),
        "E": ((3, 3), ["C", "D", "F"]),
        "F": ((3, 5), ["D", "E"])
    }, use_latlon=False)

    matcher = SimpleMatcher(mapdb, max_dist=None, min_prob_norm=None,
                                  non_emitting_states = True, only_edges=False)

    #Matching with and without timestamps signed to the points
    path_pred = matcher.match(path, unique=False)

    path = [(p1, p2, datetime.fromtimestamp(i)) for i, (p1, p2) in enumerate(path)]
    path_pred_time = matcher.match(path, unique=False)

    if directory:
        from leuvenmapmatching import visualization as mmviz
        matcher.print_lattice_stats()
        matcher.print_lattice()
        mmviz.plot_map(mapdb, matcher=matcher, show_labels=True, show_matching=True,
                       filename=str(directory / "test_nonemitting_test_path_duplicate.png"))

    # The path should be identical regardless of the timestamps
    assert path_pred == path_pred_time, f"Nodes not equal:\n{path_pred}\n{path_pred_time}"


def test_path3_many_obs():
    path = [(1, 0), (3, -0.1), (3.7, 0.6), (4.5, 0.7),
            (5.5, 1.2), (6.5, 0.88), (7.5, 0.65), (8.5, -0.1),
            (9.8, 0.1),(10.1, 1.9)]
    path_sol = ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I']
    mapdb = InMemMap("map", graph={
        "A": ((1, 0.00), ["B"]),
        "B": ((3, 0.00), ["A", "C"]),
        "C": ((4, 0.70), ["B", "D"]),
        "D": ((5, 1.00), ["C", "E"]),
        "E": ((6, 1.00), ["D", "F"]),
        "F": ((7, 0.70), ["E", "G"]),
        "G": ((8, 0.00), ["F", "H"]),
        "H": ((10, 0.0), ["G", "I"]),
        "I": ((10, 2.0), ["H"])
    }, use_latlon=False)
    matcher = SimpleMatcher(mapdb, max_dist_init=0.2, obs_noise=1, obs_noise_ne=10,
                                  non_emitting_states=True)
    matcher.match(path)
    path_pred = matcher.path_pred_onlynodes
    if directory:
        matcher.print_lattice_stats()
        matcher.print_lattice()
        from leuvenmapmatching import visualization as mmviz
        mmviz.plot_map(mapdb, matcher=matcher, show_labels=True, show_matching=True, linewidth=10,
                       show_graph=True, show_lattice=True,
                       filename=str(directory / "test_test_path_ne_3_mo.png"))
    assert path_pred == path_sol, f"Nodes not equal:\n{path_pred}\n{path_sol}"


def test_path3_few_obs_en():
    path = [(1, 0), (7.5, 0.65), (10.1, 1.9)]
    path_sol = ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I']
    mapdb = InMemMap("map", graph={
        "A": ((1, 0.00), ["B"]),
        "B": ((3, 0.00), ["A", "C"]),
        "C": ((4, 0.70), ["B", "D"]),
        "D": ((5, 1.00), ["C", "E"]),
        "E": ((6, 1.00), ["D", "F"]),
        "F": ((7, 0.70), ["E", "G"]),
        "G": ((8, 0.00), ["F", "H"]),
        "H": ((10, 0.0), ["G", "I"]),
        "I": ((10, 2.0), ["H"])
    }, use_latlon=False)
    matcher = SimpleMatcher(mapdb, max_dist_init=0.2, obs_noise=1, obs_noise_ne=10,
                                  non_emitting_states=True, only_edges=False)
    matcher.match(path)
    path_pred = matcher.path_pred_onlynodes
    if directory:
        matcher.print_lattice_stats()
        matcher.print_lattice()
        from leuvenmapmatching import visualization as mmviz
        mmviz.plot_map(mapdb, matcher=matcher, show_labels=True, show_matching=True, linewidth=10,
                       filename=str(directory / "test_test_path_ne_3_fo.png"))
    assert path_pred == path_sol, f"Nodes not equal:\n{path_pred}\n{path_sol}"


def test_path3_few_obs_e():
    path = [(1, 0), (7.5, 0.65), (10.1, 1.9)]
    path_sol = ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I']
    mapdb = InMemMap("map", graph={
        "A": ((1, 0.00), ["B"]),
        "B": ((3, 0.00), ["A", "C"]),
        "C": ((4, 0.70), ["B", "D"]),
        "D": ((5, 1.00), ["C", "E"]),
        "E": ((6, 1.00), ["D", "F"]),
        "F": ((7, 0.70), ["E", "G"]),
        "G": ((8, 0.00), ["F", "H"]),
        "H": ((10, 0.0), ["G", "I"]),
        "I": ((10, 2.0), ["H"])
    }, use_latlon=False)
    matcher = SimpleMatcher(mapdb, max_dist_init=0.2, obs_noise=1, obs_noise_ne=10,
                                  non_emitting_states=True, only_edges=True)
    matcher.match(path)
    path_pred = matcher.path_pred_onlynodes
    if directory:
        matcher.print_lattice_stats()
        matcher.print_lattice()
        from leuvenmapmatching import visualization as mmviz
        mmviz.plot_map(mapdb, matcher=matcher, show_labels=True, show_matching=True, linewidth=10,
                       filename=str(directory / "test_test_path_e_3_fo.png"))
    assert path_pred == path_sol, f"Nodes not equal:\n{path_pred}\n{path_sol}"


def test_path3_dist():
    path = [(0, 1), (0.65, 7.5), (1.9, 10.1)]
    path_sol = ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I']
    mapdb = InMemMap("map", graph={
        "A": ((0.00, 1), ["B"]),
        "B": ((0.00, 3), ["A", "C"]),
        "C": ((0.70, 3), ["B", "D"]),
        "D": ((1.00, 5), ["C", "E"]),
        "E": ((1.00, 6), ["D", "F"]),
        "F": ((0.70, 7), ["E", "G"]),
        "G": ((0.00, 8), ["F", "H"]),
        "H": ((0.0, 10), ["G", "I"]),
        "I": ((2.0, 10), ["H"])
    }, use_latlon=False)
    matcher = DistanceMatcher(mapdb, max_dist_init=0.2,
                              obs_noise=0.5, obs_noise_ne=2, dist_noise=0.5,
                              non_emitting_states=True)
    states, lastidx = matcher.match(path)
    path_pred = matcher.path_pred_onlynodes
    if directory:
        from leuvenmapmatching import visualization as mmviz
        mmviz.plot_map(mapdb, matcher=matcher, show_labels=True, show_matching=True, linewidth=2,
                       filename=str(directory / "test_path_3_dist.png"))
    assert path_pred == path_sol, f"Nodes not equal:\n{path_pred}\n{path_sol}"

    for obs_idx, m in enumerate(matcher.lattice_best):  # type: Tuple[int, DistanceMatching]
        state = m.shortkey  # tuple indicating edge
        ne_str = "e" if m.is_emitting() else "ne"  # state is emitting or not
        p1_str = "{:>5.2f}-{:<5.2f}".format(*m.edge_m.pi)  # best matching location on graph
        p2_str = "{:>5.2f}-{:<5.2f}".format(*m.edge_o.pi)  # best matching location on track
        print(f"{obs_idx:<2} | {state} | {ne_str:<2} | {p1_str} | {p2_str}")


if __name__ == "__main__":
    # mm.matching.logger.setLevel(logging.INFO)
    logger.setLevel(logging.DEBUG)
    logger.addHandler(logging.StreamHandler(sys.stdout))
    directory = Path(os.environ.get('TESTDIR', Path(__file__).parent))
    print(f"Saving files to {directory}")
    # visualize_map(pathnb=1)
    # test_path1()
    # test_path1_inc()
    # test_path1_dist()
    test_path2()
    # test_path2_dist()
    # test_path2_incremental()
    # test_path_duplicate()
    # test_path3_many_obs()
    # test_path3_few_obs_en()
    # test_path3_few_obs_e()
    # test_path3_dist()



# encoding: utf-8
"""
leuvenmapmatching.matcher.simple
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

:author: Wannes Meert
:copyright: Copyright 2015-2018 DTAI, KU Leuven and Sirris.
:license: Apache License, Version 2.0, see LICENSE for details.
"""
import math
from scipy.stats import halfnorm, norm

from .base import BaseMatcher, BaseMatching
from ..util.segment import Segment


class SimpleMatching(BaseMatching):
    pass


class SimpleMatcher(BaseMatcher):

    def __init__(self, *args, **kwargs):
        """A simple matcher that prefers paths where each matched location is as close as possible to the
        observed position.

        :param avoid_goingback: Change the transition probability to be lower for the direction the path is coming
            from.
        :param kwargs: Arguments passed to :class:`BaseMatcher`.
        """
        if "matching" not in kwargs:
            kwargs['matching'] = SimpleMatching
        super().__init__(*args, **kwargs)

        self.obs_noise_dist = halfnorm(scale=self.obs_noise)
        self.obs_noise_dist_ne = halfnorm(scale=self.obs_noise_ne)
        # normalize to max 1 to simulate a prob instead of density
        self.obs_noise_logint = math.log(self.obs_noise * math.sqrt(2 * math.pi) / 2)
        self.obs_noise_logint_ne = math.log(self.obs_noise_ne * math.sqrt(2 * math.pi) / 2)

        # Transition probability is divided (in logprob_trans) by this factor if we move back on the
        # current edge.
        self.avoid_goingback = kwargs.get('avoid_goingback', True)
        self.gobackonedge_factor_log = math.log(0.99)
        # Transition probability is divided (in logprob_trans) by this factor if the next state is
        # also the previous state, thus if we go back
        self.gobacktoedge_factor_log = math.log(0.5)
        # Transition probability is divided (in logprob_trans) by this factor if a transition is made
        # This is to try to stay on the same node unless there is a good reason
        self.transition_factor = math.log(0.9)

    def logprob_trans(self, prev_m: BaseMatching, edge_m: Segment, edge_o: Segment,
                      is_prev_ne=False, is_next_ne=False):
        """Transition probability.

        Note: In contrast with a regular HMM, this is not a probability density function, it needs
              to be a proper probability (thus values between 0.0 and 1.0).
        """
        logprob = 0
        if prev_m.edge_m.label == edge_m.label:
            # Staying in same state
            if self.avoid_goingback and edge_m.key == prev_m.edge_m.key and edge_m.ti < prev_m.edge_m.ti:
                # Going back on edge
                logprob += self.gobackonedge_factor_log  # prefer not going back
        else:
            # Moving states
            logprob += self.transition_factor
            if self.avoid_goingback:
                # Goin back on state
                going_back = False
                for m in prev_m.prev:
                    if edge_m.label == m.edge_m.label:
                        going_back = True
                        break
                if going_back:
                    logprob += self.gobacktoedge_factor_log  # prefer not going back
        return logprob, {}  # All probabilities are 1 (thus technically not a distribution)

    def logprob_obs(self, dist, prev_m, new_edge_m, new_edge_o, is_ne=False):
        """Emission probability.

        Note: In contrast with a regular HMM, this is not a probability density function, it needs
              to be a proper probability (thus values between 0.0 and 1.0).
        """
        if is_ne:
            result = self.obs_noise_dist_ne.logpdf(dist) + self.obs_noise_logint_ne
        else:
            result = self.obs_noise_dist.logpdf(dist) + self.obs_noise_logint
        # print("logprob_obs: {} -> {:.5f} = {:.5f}".format(dist, result, math.exp(result)))
        return result, {}



# encoding: utf-8
"""
leuvenmapmatching.matcher.base
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Base Matcher and Matching classes.

This a generic base class to be used by matchers. This class itself
does not implement a working matcher.

:author: Wannes Meert
:copyright: Copyright 2015-2021 DTAI, KU Leuven and Sirris.
:license: Apache License, Version 2.0, see LICENSE for details.
"""
from __future__ import print_function

import math
import sys
import logging
import time
from collections import OrderedDict, defaultdict, namedtuple
from itertools import islice
from typing import List, Tuple, Dict, Any, Optional, Set

import numpy as np

from ..util.segment import Segment
from ..util import approx_equal, approx_leq


logger = logging.getLogger("be.kuleuven.cs.dtai.mapmatching")
approx_value = 0.0000000001
ema_const = namedtuple('EMAConst', ['prev', 'cur'])(0.7, 0.3)
default_label_width = 25


class BaseMatching(object):
    """Matching object that represents a node in the Viterbi lattice."""
    __slots__ = ['matcher', 'edge_m', 'edge_o',
                 'logprob', 'logprobema', 'logprobe', 'logprobne',
                 'obs', 'obs_ne', 'dist_obs',
                 'prev', 'prev_other', 'stop', 'length', 'delayed']

    def __init__(self, matcher: 'BaseMatcher', edge_m: Segment, edge_o: Segment,
                 logprob=-np.inf, logprobema=-np.inf, logprobe=-np.inf, logprobne=-np.inf,
                 dist_obs: float = 0.0, obs: int = 0, obs_ne: int = 0,
                 prev: Optional[Set['BaseMatching']] = None, stop: bool = False, length: int = 1,
                 delayed: int = 0, **_kwargs):
        """

        :param matcher: Reference to the Matcher used to generate this matching object.
        :param edge_m: Segment in the given graph (thus line between two nodes in the graph).
        :param edge_o: Segment in the given observations (thus line in between two observations).
        :param logprob: Log probability of this matching.
        :param logprobema: Exponential Mean Average of Log probability.
        :param logprobe: Emitting
        :param logprobne: Non-emitting
        :param dist_obs: Distance between map point and observation
        :param obs: Reference to path entry index (observation)
        :param obs_ne: Number of non-emitting states for this observation
        :param prev: Previous best matching objects
        :param stop: Stop after this matching (e.g. because probability is too low)
        :param length: Lenght of current matching sequence through lattice.
        :param delayed: This matching is temporarily stopped if >0 (e.g. to first explore better options).
        :param dist_m: Distance over graph
        :param dist_o: Distance over observations
        :param _kwargs:
        """
        self.edge_m: Segment = edge_m
        self.edge_o: Segment = edge_o
        self.logprob: float = logprob        # max probability
        self.logprobe: float = logprobe      # Emitting
        self.logprobne: float = logprobne    # Non-emitting
        self.logprobema: float = logprobema  # exponential moving average log probability  # TODO: Not used anymore?
        self.obs: int = obs  # reference to path entry index (observation)
        self.obs_ne: int = obs_ne  # number of non-emitting states for this observation
        self.dist_obs: float = dist_obs  # Distance between map point and observation
        self.prev: Set[BaseMatching] = set() if prev is None else prev  # Previous best matching objects
        self.prev_other: Set[BaseMatching] = set()  # Previous matching objects with lower logprob
        self.stop: bool = stop
        self.length: int = length
        self.delayed: int = delayed
        self.matcher: BaseMatcher = matcher

    @property
    def prune_value(self):
        """Pruning the lattice (e.g. to delay) is based on this key."""
        return self.logprob
        # return self.logprobema

    def next(self, edge_m: Segment, edge_o: Segment, obs: int = 0, obs_ne: int = 0):
        """Create a next lattice Matching object with this Matching object as the previous one in the lattice."""
        new_stop = False
        if edge_m.is_point() and edge_o.is_point():
            # node to node
            dist = self.matcher.map.distance(edge_m.p1, edge_o.p1)
            # proj_m = edge_m.p1
            # proj_o = edge_o.pi
        elif edge_m.is_point() and not edge_o.is_point():
            # node to edge
            dist, proj_o, t_o = self.matcher.map.distance_point_to_segment(edge_m.p1, edge_o.p1, edge_o.p2)
            # proj_m = edge_m.p1
            edge_o.pi = proj_o
            edge_o.ti = t_o
        elif not edge_m.is_point() and edge_o.is_point():
            # edge to node
            dist, proj_m, t_m = self.matcher.map.distance_point_to_segment(edge_o.p1, edge_m.p1, edge_m.p2)
            if not self.matcher.only_edges and (approx_equal(t_m, 0.0) or approx_equal(t_m, 1.0)):
                if __debug__ and logger.isEnabledFor(logging.DEBUG):
                    logger.debug(f"   | Stopped trace: Too close to end, {t_m}")
                    new_stop = True
                else:
                    return None
            edge_m.pi = proj_m
            edge_m.ti = t_m
            # proj_o = edge_o.pi
        elif not edge_m.is_point() and not edge_o.is_point():
            # edge to edge
            dist, proj_m, proj_o, t_m, t_o = self.matcher.map.distance_segment_to_segment(edge_m.p1, edge_m.p2,
                                                                                          edge_o.p1, edge_o.p2)
            edge_m.pi = proj_m
            edge_m.ti = t_m
            edge_o.pi = proj_o
            edge_o.ti = t_o
        else:
            raise Exception(f"Should not happen")

        logprob_trans, props_trans = self.matcher.logprob_trans(self, edge_m, edge_o,
                                                                is_prev_ne=(self.obs_ne != 0),
                                                                is_next_ne=(obs_ne != 0))
        logprob_obs, props_obs = self.matcher.logprob_obs(dist, self, edge_m, edge_o,
                                                          is_ne=(obs_ne != 0))
        if __debug__ and logprob_trans > 0:
            raise Exception(f"logprob_trans = {logprob_trans} > 0")
        if __debug__ and logprob_obs > 0:
            raise Exception(f"logprob_obs = {logprob_obs} > 0")
        new_logprob_delta = logprob_trans + logprob_obs
        if obs_ne == 0:
            new_logprobe = self.logprob + new_logprob_delta
            new_logprobne = 0
            new_logprob = new_logprobe
            new_length = self.length + 1
        else:
            # Non-emitting states require normalisation
            # "* e^(ne_length_factor_log)" or "- ne_length_factor_log" for every step to a non-emitting
            # state to prefer shorter paths
            new_logprobe = self.logprobe + self.matcher.ne_length_factor_log
            # The obvious choice would be average to compensate for that non-emitting states
            # create different path lengths between emitting nodes.
            # We use min() as it is a monotonic function, in contrast with an average
            new_logprobne = min(self.logprobne, new_logprob_delta)
            new_logprob = new_logprobe + new_logprobne
            # Alternative approach with an average
            # new_logprobne = self.logprobne + new_logprob_delta
            # "+ 1" to punish non-emitting states a bit less. Otherwise it would be
            # similar to (Pr_tr*Pr_obs)**2, which punishes just one non-emitting state too much.
            # new_logprob = new_logprobe + new_logprobne / (obs_ne + 1)
            new_length = self.length
        new_logprobema = ema_const.cur * new_logprob_delta + ema_const.prev * self.logprobema
        new_stop |= self.matcher.do_stop(new_logprob / new_length, dist, logprob_trans, logprob_obs)
        if __debug__ and new_logprob > self.logprob:
            raise Exception(f"Expecting a monotonic probability, "
                            f"new_logprob = {new_logprob} > logprob = {self.logprob}")
        if not new_stop or (__debug__ and logger.isEnabledFor(logging.DEBUG)):
            m_next = self.__class__(self.matcher, edge_m, edge_o,
                                    logprob=new_logprob, logprobne=new_logprobne,
                                    logprobe=new_logprobe, logprobema=new_logprobema,
                                    obs=obs, obs_ne=obs_ne, prev={self}, dist_obs=dist,
                                    stop=new_stop, length=new_length, delayed=self.delayed,
                                    **props_trans, **props_obs)
            return m_next
        else:
            return None

    @classmethod
    def first(cls, logprob_init, edge_m, edge_o, matcher, dist_obs):
        """Create an initial lattice Matching object."""
        logprob_obs, props_obs = matcher.logprob_obs(dist_obs, None, edge_m, edge_o)
        logprob = logprob_init + logprob_obs
        new_stop = matcher.do_stop(logprob, dist_obs, logprob_init, logprob_obs)
        if not new_stop or logger.isEnabledFor(logging.DEBUG):
            m_next = cls(matcher, edge_m=edge_m, edge_o=edge_o,
                         logprob=logprob, logprobema=logprob, logprobe=logprob, logprobne=0,
                         dist_obs=dist_obs, obs=0, stop=new_stop, **props_obs)
            return m_next
        else:
            return None

    def update(self, m_next):
        """Update the current entry if the new matching object for this state is better.

        :param m_next: The new matching object representing the same node in the lattice.
        :return: True if the current object is replaced, False otherwise
        """
        # if self.length != m_next.length:
        #     slogprob_norm = self.logprob / self.length
        #     nlogprob_norm = m_next.logprob / m_next.length
        # else:
        #     slogprob_norm = self.logprob
        #     nlogprob_norm = m_next.logprob
        # if (self.stop == m_next.stop and slogprob_norm < nlogprob_norm) or (self.stop and not m_next.stop):
        #     self._update_inner(m_next)
        #     return True
        # elif abs(slogprob_norm - nlogprob_norm) < approx_value and self.stop == m_next.stop:
        #     self.prev.update(m_next.prev)
        #     self.stop = m_next.stop
        #     return False
        assert self.length == m_next.length
        if (self.stop and not m_next.stop) \
                or (self.stop == m_next.stop and self.logprob < m_next.logprob):
            self._update_inner(m_next)
            return True
        else:
            self.prev_other.update(m_next.prev)
            return False

    def _update_inner(self, m_other: 'BaseMatching'):
        self.edge_m = m_other.edge_m
        self.edge_o = m_other.edge_o
        self.logprob = m_other.logprob
        self.logprobe = m_other.logprobe
        self.logprobne = m_other.logprobne
        self.logprobema = m_other.logprobema
        self.dist_obs = m_other.dist_obs
        self.obs = m_other.obs
        self.obs_ne = m_other.obs_ne
        self.prev_other.update(self.prev)  # Do we use this?
        self.prev = m_other.prev
        self.stop = m_other.stop
        self.delayed = m_other.delayed
        self.length = m_other.length

    def is_nonemitting(self):
        return self.obs_ne != 0

    def is_emitting(self):
        return self.obs_ne == 0

    def last_emitting_logprob(self):
        if self.is_emitting():
            return self.logprob
        elif self.prev is None or len(self.prev) == 0:
            return 0
        else:
            return next(iter(self.prev)).last_emitting_logprob()

    def __str__(self, label_width=None):
        stop = ''
        if self.stop:
            stop = 'x'
        else:
            stop = f'{self.delayed}'
        if label_width is None:
            label_width = default_label_width
        repr_tmpl = "{:<2} | {:<"+str(label_width)+"} | {:10.5f} | {:10.5f} | {:10.5f} | {:10.5f} | " +\
                    "{:<3} | {:10.5f} | {:<" + str(label_width) + "} |"
        return repr_tmpl.format(stop, self.label, self.logprob, self.logprob / self.length,
                                self.logprobema, self.logprobne, self.obs,
                                self.dist_obs, ",".join([str(prev.label) for prev in self.prev]))

    def __repr__(self):
        return "Matching<"+str(self.label)+">"

    @staticmethod
    def repr_header(label_width=None, stop=""):
        if label_width is None:
            label_width = default_label_width
        repr_tmpl = "{:<2} | {:<"+str(label_width)+"} | {:<10} | {:<10} | {:<10} | {:<10} | " + \
                    "{:<3} | {:<10} | {:<"+str(label_width)+"} |"
        return repr_tmpl.format(stop, "", "lg(Pr)", "nlg(Pr)", "slg(Pr)", "lg(Pr-ne)", "obs", "d(obs)", "prev")

    @staticmethod
    def repr_static(fields, label_width=None):
        if label_width is None:
            label_width = default_label_width
        default_fields = ["", "", float('nan'), float('nan'), float('nan'), float('nan'), "", float('nan'), "", ""]
        repr_tmpl = "{:<2} | {:<" + str(label_width) + "} | {:10.5f} | {:10.5f} | {:10.5f} | {:10.5f} | " + \
                    "{:<3} | {:10.5f} | {:<" + str(label_width) + "} |"
        if len(fields) < 8:
            fields = list(fields) + default_fields[len(fields):]
        return repr_tmpl.format(*fields)

    @property
    def label(self):
        if self.edge_m.p2 is None:
            return "{}---{}-{}".format(self.edge_m.l1, self.obs, self.obs_ne)
        else:
            return "{}-{}-{}-{}".format(self.edge_m.l1, self.edge_m.l2, self.obs, self.obs_ne)

    @property
    def cname(self):
        if self.edge_m.l2 is None:
            return "{}_{}_{}".format(self.edge_m.l1, self.obs, self.obs_ne)
        else:
            return "{}_{}_{}_{}".format(self.edge_m.l1, self.edge_m.l2, self.obs, self.obs_ne)

    @property
    def key(self):
        """Key that indicates the node or edge, observation and non-emitting step.
        This is the unique key that is used in the lattice.
        """
        if self.edge_m.l2 is None:
            return tuple([self.edge_m.l1, self.obs, self.obs_ne])
        else:
            return tuple([self.edge_m.l1, self.edge_m.l2, self.obs, self.obs_ne])

    @property
    def shortkey(self):
        """Key that indicates the node or edge. Irrespective of the current observation."""
        if self.edge_m.l2 is None:
            return self.edge_m.l1
        else:
            return tuple([self.edge_m.l1, self.edge_m.l2])

    @property
    def nodes(self):
        if self.edge_m.l2 is None:
            return [self.edge_m.l1]
        else:
            return [self.edge_m.l1, self.edge_m.l2]

    def __hash__(self):
        return self.cname.__hash__()


class LatticeColumn:

    def __init__(self, obs_idx):
        # 0 = obs, >0 = non-emitting between this obs and next
        self.obs_idx = obs_idx
        self.o = []  # type list[dict[label,Matching]]

    def __contains__(self, item):
        for c in self.o:
            if item in c:
                return True
        return False

    def __len__(self):
        return len(self.o)

    def set_delayed(self, delayed):
        """Update all delayed values."""
        for c in self.o:
            for m in c.values():
                m.delayed = delayed

    def dict(self, obs_ne=None):
        if obs_ne is None:
            raise AttributeError('obs_ne should be value')
        while obs_ne >= len(self.o):
            self.o.append({})
        return self.o[obs_ne]

    def values_all(self):
        """All matches for the emitting layer and all non-emitting layers."""
        values = set()
        for o in self.o:
            values.update(o.values())
        return values

    def values(self, obs_ne=None):
        if obs_ne is None:
            raise AttributeError('obs_ne should be value')
        if len(self.o) <= obs_ne:
            return []
        return self.o[obs_ne].values()

    def upsert(self, matching):
        # type: (BaseMatching) -> None
        if matching is None:
            return None
        while matching.obs_ne >= len(self.o):
            self.o.append({})
        c = self.o[matching.obs_ne]
        if matching.key in c:
            other_matching = c[matching.key]  # type: BaseMatching
            other_matching.update(matching)
        else:
            c[matching.key] = matching
        return c[matching.key]

    def prune(self, obs_ne, max_lattice_width, expand_upto, prune_thr=None):
        """Prune given column in the lattice to fit in max_lattice_width.
        Also ignore all matchings with a probability lower than prune_thr. These are
        matchings that are worse than the matchings at the next observation that are
        retained after pruning.

        :param obs_ne:
        :param max_lattice_width:
        :param expand_upto: The current expand level
        :return:
        """
        cur_lattice = [m for m in self.values(obs_ne) if not m.stop]
        if __debug__:
            logger.debug('Prune lattice[{},{}] from {} to {}, with prune thr {}'
                         .format(self.obs_idx, obs_ne,
                                 len([m for m in cur_lattice if not m.stop and m.delayed == expand_upto]),
                                 max_lattice_width, prune_thr))
            cnt_pruned = 0
        if max_lattice_width is not None and len(cur_lattice) > max_lattice_width:
            ms = sorted(cur_lattice, key=lambda t: t.prune_value, reverse=True)
            cur_width = max_lattice_width
            m_last = ms[cur_width - 1]
            # Extend current width if next pruned matching has same logprob as last kept matching
            # This increases the lattice width but otherwise the algorithm depends on the
            # order of edges/nodes and is not deterministic.
            while cur_width < len(ms) and ms[cur_width].prune_value == m_last.prune_value:
                m_last = ms[cur_width]
                cur_width += 1
            if prune_thr is not None:
                while cur_width > 0 and ms[cur_width - 1].prune_value < prune_thr:
                    cur_width -= 1
            for m in ms[:cur_width]:  # type: BaseMatching
                if m.delayed > expand_upto:
                    m.delayed = expand_upto  # expand now
            for m in ms[cur_width:]:
                if m.delayed <= expand_upto:
                    if __debug__:
                        cnt_pruned += 1
                    m.delayed = expand_upto + 1  # expand later
            if cur_width > 0:
                prune_thr = ms[cur_width - 1].prune_value
        if __debug__:
            logger.debug(f'Pruned {cnt_pruned} matchings, return {prune_thr}')
        return prune_thr


class BaseMatcher:

    def __init__(self, map_con, obs_noise=1, max_dist_init=None, max_dist=None, min_prob_norm=None,
                 non_emitting_states=True, max_lattice_width=None,
                 only_edges=True, obs_noise_ne=None, matching=BaseMatching,
                 non_emitting_length_factor=0.75, **kwargs):
        """Initialize a matcher for map matching.

        This a generic base class to be used by matchers. This class itself
        does not implement a working matcher.

        Distances are in meters when using latitude-longitude.

        :param map_con: Map object to connect to map database
        :param obs_noise: Standard deviation of noise
        :param obs_noise_ne: Standard deviation of noise for non-emitting states (is set to obs_noise if not give)
        :param max_dist_init: Maximum distance from start location (if not given, uses max_dist)
        :param max_dist: Maximum distance from path (this is a hard cut, min_prob_norm should be better)
        :param min_prob_norm: Minimum normalized probability of observations (ema)
        :param non_emitting_states: Allow non-emitting states. A non-emitting state is a state that is
            not associated with an observation. Here we assume it can be associated with a location in between
            two observations to allow for pruning. It is advised to set min_prob_norm and/or max_dist to avoid
            visiting all possible nodes in the graph.
        :param max_lattice_width: Only continue from a limited number of states (thus locations) for a given observation.
            This possibly speeds up the matching by a lot.
            If there are more possible next states, the states with the best likelihood so far are selected.
            The other states are 'delayed'. If the matching is continued later with a larger value using
            `increase_max_lattice_width`, the algorithms continuous from these delayed states.
        :param only_edges: Do not include nodes as states, only edges. This is the typical setting for HMM methods.
        :param matching: Matching type
        :param non_emitting_length_factor: Reduce the probability of a sequence of non-emitting states the longer it
            is. This can be used to prefer shorter paths. This is separate from the transition probabilities because
            transition probabilities are averaged for non-emitting states and thus the length is also averaged out.

        To define a custom transition and/or emission probability distribtion, overwrite the following functions:

        - :meth:`logprob_trans`
        - :meth:`logprob_obs`

        """
        self.map = map_con  # type: BaseMap
        if max_dist:
            self.max_dist = max_dist
        else:
            self.max_dist = np.inf
        if max_dist_init:
            self.max_dist_init = max_dist_init
        else:
            self.max_dist_init = self.max_dist
        if min_prob_norm:
            self.min_logprob_norm = math.log(min_prob_norm)
        else:
            self.min_logprob_norm = -np.inf
        logger.debug(f"Matcher.min_logprob_norm = {self.min_logprob_norm}, Matcher.max_dist = {self.max_dist}")
        self.obs_noise = obs_noise
        if obs_noise_ne is None:
            self.obs_noise_ne = obs_noise
        else:
            self.obs_noise_ne = obs_noise_ne

        self.path = None
        self.lattice = None  # type: Optional[dict[int,LatticeColumn]]
        # Best path through lattice:
        self.lattice_best = None  # type: Optional[list[BaseMatching]]
        self.node_path = None  # type: Optional[list[str]]
        self.matching = matching
        self.non_emitting_states = non_emitting_states  # type: bool
        self.non_emitting_states_maxnb = 100
        self.max_lattice_width = max_lattice_width  # type: Optional[int]
        self.only_edges = only_edges  # type: bool
        self.expand_now = 0  # all m.delayed <= expand_upto will be expanded

        # Penalties
        self.ne_length_factor_log = math.log(non_emitting_length_factor)

    def logprob_trans(self, prev_m, edge_m, edge_o,
                      is_prev_ne=False, is_next_ne=False):
        # type: (BaseMatcher, BaseMatching, Segment, Segment, bool, bool) -> Tuple[float, Dict[str, Any]]
        """Transition probability.

        Note: In contrast with a regular HMM, this cannot be a probability density function, it needs
              to be a proper probability (thus values between 0.0 and 1.0).

        :return: probability, properties that are passed to the matching object
        """
        return 0, {}  # All probabilities are 1 (thus technically not a distribution)

    def logprob_obs(self, dist, prev_m, new_edge_m, new_edge_o, is_ne=False):
        """Emission probability.

        Note: In contrast with a regular HMM, this cannot be a probability density function, it needs
              to be a proper probability (thus values between 0.0 and 1.0).

        :return: probability, properties that are passed to the matching object
        """
        return 0, {}

    def match_gpx(self, gpx_file, unique=True):
        """Map matching from a gpx file"""
        from ..util.gpx import gpx_to_path
        path = gpx_to_path(gpx_file)
        return self.match(path, unique=unique)

    def do_stop(self, logprob_norm, dist, logprob_trans, logprob_obs):
        if logprob_norm < self.min_logprob_norm:
            logger.debug(f"   | Stopped trace: norm(log(Pr)) too small: {logprob_norm} < {self.min_logprob_norm}"
                         f"  -- lPr_t = {logprob_trans:.3f}, lPr_o = {logprob_obs:.3f}")
            return True
        if dist > self.max_dist:
            logger.debug(f"   | Stopped trace: distance too large: {dist} > {self.max_dist}")
            return True
        return False

    def _insert(self, m_next):
        return self.lattice[m_next.obs].upsert(m_next)

    def match(self, path, unique=False, tqdm=None, expand=False):
        """Dynamic Programming based (HMM-like) map matcher.

        If the matcher fails to match the entire path, the last matched index is returned.
        This index can be used to run the matcher again from that observation onwards.

        :param path: list[Union[tuple[lat, lon], tuple[lat, lon, time]]
        :param unique: Only retain unique nodes in the sequence (avoid repetitions)
        :param tqdm: Use a tqdm progress reporter (default is None)
        :param expand: Expand the current lattice (delayed matches)
        :return: Tuple of (List of BaseMatching, index of last observation that was matched)
        """
        if __debug__:
            logger.debug("Start matching path of length {}".format(len(path)))

        # Initialisation
        if expand:
            self.expand_now += 1
            if self.path != path:
                is_path_extended = True
                if len(path) > len(self.path):
                    for pi, spi in zip(path, self.path):
                        if pi != spi:
                            is_path_extended = False
                            break
                else:
                    is_path_extended = False
                if is_path_extended:
                    self.lattice[len(self.path) - 1].set_delayed(self.expand_now)
                    for obs_idx in range(len(self.path), len(path)):
                        if obs_idx not in self.lattice:
                            self.lattice[obs_idx] = LatticeColumn(obs_idx)
                    self.path = path
                else:
                    raise Exception(f'Cannot expand for a new path, should be the same path (or an extension).')
        else:
            self.path = path
            self.expand_now = 0

        nb_start_nodes = self._create_start_nodes(use_edges=self.only_edges)
        if nb_start_nodes == 0:
            self.lattice_best = []
            return [], 0
        if __debug__ and logger.isEnabledFor(logging.DEBUG):
            self.print_lattice(obs_idx=0, label_width=default_label_width, debug=True)

        # Start iterating over observations 1..end
        t_start = time.time()
        iterator = range(1, len(path))
        if tqdm:
            iterator = tqdm(iterator)
        early_stop_idx = None
        for obs_idx in iterator:
            if __debug__:
                logger.debug("--- obs {} --- {} ---".format(obs_idx, self.path[obs_idx]))
            # check if early stopping has occured
            cnt_lat_size_not_zero = False
            for m_tmp in self.lattice[obs_idx - 1].values(0):
                if not m_tmp.stop:
                    cnt_lat_size_not_zero = True
                    break
            # if len(self.lattice[obs_idx - 1]) == 0:
            if not cnt_lat_size_not_zero:
                if __debug__:
                    logger.debug("No solutions found anymore")
                early_stop_idx = obs_idx - 1
                logger.info(f'Stopped early at observation {early_stop_idx}')
                break
            # Expand matches
            self._match_states(obs_idx)
            if self.non_emitting_states:
                # Fill in non-emitting states between previous and current observation
                self._match_non_emitting_states(obs_idx - 1, expand=expand)
            if self.max_lattice_width:
                # Prune again if non_emitting_states reactives matches from match_states
                self.lattice[obs_idx].prune(0, self.max_lattice_width, self.expand_now)
            if __debug__ and logger.isEnabledFor(logging.DEBUG):
                self.print_lattice(obs_idx=obs_idx, label_width=default_label_width, debug=True)
                logger.debug(f"--- end obs {obs_idx} ---")

        t_delta = time.time() - t_start
        logger.info("--- end ---")
        logger.info("Build lattice in {} seconds".format(t_delta))

        # Backtrack to find best path
        if not early_stop_idx:
            one_no_stop = False
            for m in self.lattice[len(path) - 1].values_all():  # todo: could be values(0) ?
                if not m.stop:
                    one_no_stop = True
                    break
            if not one_no_stop:
                early_stop_idx = len(path) - 1
        if early_stop_idx is not None:
            if early_stop_idx == 0:
                self.lattice_best = []
                return [], 0
            start_idx = early_stop_idx - 1
        else:
            start_idx = len(self.path) - 1
        node_path = self._build_node_path(start_idx, unique)
        return node_path, start_idx

    def _skip_ne_states(self, prev_m):
        # type: (BaseMatcher, BaseMatching) -> bool
        return False

    def _create_start_nodes(self, use_edges=True):
        """Find those nodes that are close to the first point in the path.

        :return: Number of created start points.
        """
        # Initialisation on first observation
        if self.expand_now > 0:
            # No need to search for new points, only activate delayed matches
            self.lattice[0].prune(0, self.max_lattice_width, self.expand_now)
            return len(self.lattice[0])

        t_start = time.time()
        self.lattice = dict()
        for obs_idx in range(len(self.path)):
            self.lattice[obs_idx] = LatticeColumn(obs_idx)

        if use_edges:
            nodes = self.map.edges_closeto(self.path[0], max_dist=self.max_dist_init)
        else:
            nodes = self.map.nodes_closeto(self.path[0], max_dist=self.max_dist_init)
        if __debug__:
            logger.debug("--- obs {} --- {} ---".format(0, self.path[0]))
        t_delta = time.time() - t_start
        logger.info("Initialized lattice with {} starting points in {} seconds".format(len(nodes), t_delta))
        if len(nodes) == 0:
            logger.info(f'Stopped early at observation 0'
                        f', no starting points/edges x found for which '
                        f'|x - ({self.path[0][0]:.2f},{self.path[0][1]:.2f})| < {self.max_dist_init}')
            return 0
        if __debug__:
            logger.debug(self.matching.repr_header())
        logprob_init = 0  # math.log(1.0/len(nodes))
        if use_edges:
            # Search for nearby edges
            for dist_obs, label1, loc1, label2, loc2, pi, ti in nodes:
                if label2 == label1:
                    continue
                edge_m = Segment(label1, loc1, label2, loc2, pi, ti)
                edge_o = Segment(f"O{0}", self.path[0])
                m_next = self.matching.first(logprob_init, edge_m, edge_o, self, dist_obs)
                if m_next is not None:
                    self.lattice[0].upsert(m_next)
                    if __debug__:
                        logger.debug(str(m_next))
        else:
            # Search for nearby nodes
            for dist_obs, label, loc in nodes:
                edge_m = Segment(label, loc)
                edge_o = Segment(f"O{0}", self.path[0])
                m_next = self.matching.first(logprob_init, edge_m, edge_o, self, dist_obs)
                if m_next is not None:
                    self.lattice[0].upsert(m_next)
                    if __debug__:
                        logger.debug(str(m_next))
        if self.max_lattice_width:
            self.lattice[0].prune(0, max_lattice_width=self.max_lattice_width, expand_upto=self.expand_now)
            # if self.non_emitting_states:
            #     self._match_non_emitting_states(0, path)
        return len(self.lattice[0])

    def _match_states(self, obs_idx):
        """Match states

        :param obs_idx:
        :return: True is new states have been found, False otherwise.
        """
        prev_lattice = [m for m in self.lattice[obs_idx - 1].values(0) if not m.stop and m.delayed == self.expand_now]
        count = 0
        for m in prev_lattice:  # type: BaseMatching
            if m.stop:
                assert(False)  # should not happen
                continue
            count += 1
            if m.edge_m.is_point():
                # == Move to neighbour from node ==
                nbrs = self.map.nodes_nbrto(m.edge_m.l1)
                # print("Neighbours for {}: {}".format(m, nbrs))
                if nbrs is None:
                    if __debug__:
                        logger.debug("No neighbours found for node {}".format(m.edge_m.l1))
                    continue
                if __debug__:
                    logger.debug("   + Move to {} neighbours from node {}".format(len(nbrs), m.edge_m.l1))
                    logger.debug(m.repr_header())
                for nbr_label, nbr_loc in nbrs:
                    # === Move from node to node (or stay on node) ===
                    if not self.only_edges:
                        edge_m = Segment(nbr_label, nbr_loc)
                        edge_o = Segment(f"O{obs_idx}", self.path[obs_idx])
                        m_next = m.next(edge_m, edge_o, obs=obs_idx)
                        if m_next is not None:
                            self._insert(m_next)
                            if __debug__:
                                logger.debug(str(m_next))

                    # === Move from node to edge ===
                    if m.edge_m.l1 != nbr_label:
                        edge_m = Segment(m.edge_m.l1, m.edge_m.p1, nbr_label, nbr_loc)
                        edge_o = Segment(f"O{obs_idx}", self.path[obs_idx])
                        m_next = m.next(edge_m, edge_o, obs=obs_idx)
                        if m_next is not None:
                            self._insert(m_next)
                            if __debug__:
                                logger.debug(str(m_next))
                    else:
                        if __debug__:
                            logger.debug(self.matching.repr_static(('x', f'{nbr_label}-{nbr_label} < self-loop')))

            else:
                # == Move to neighbour from edge ==
                if __debug__:
                    logger.debug("   + Move to neighbour from edge {}".format(m.label))
                    logger.debug(m.repr_header())

                # === Stay on edge ===
                edge_m = Segment(m.edge_m.l1, m.edge_m.p1, m.edge_m.l2, m.edge_m.p2)
                edge_o = Segment(f"O{obs_idx}", self.path[obs_idx])
                m_next = m.next(edge_m, edge_o, obs=obs_idx)
                if m_next is not None:
                    self._insert(m_next)
                    if __debug__:
                        logger.debug(str(m_next))

                # === Move from edge to node ===
                if not self.only_edges:
                    edge_m = Segment(m.edge_m.l2, m.edge_m.p2)
                    edge_o = Segment(f"O{obs_idx}", self.path[obs_idx])
                    m_next = m.next(edge_m, edge_o, obs=obs_idx)
                    if m_next is not None:
                        self._insert(m_next)
                        if __debug__:
                            logger.debug(str(m_next))

                else:
                    # === Move from edge to next edge ===
                    nbrs = self.map.edges_nbrto((m.edge_m.l1, m.edge_m.l2))
                    if nbrs is None or len(nbrs) == 0:
                        if __debug__:
                            logger.debug(f"No neighbours found for edge {m.edge_m.label}")
                        continue
                    for nbr_label1, nbr_loc1, nbr_label2, nbr_loc2 in nbrs:
                        # same edge is different action, opposite edge should be allowed to return in a one-way street
                        if m.edge_m.l2 != nbr_label2 and m.edge_m.l1 != nbr_label1:
                            edge_m = Segment(nbr_label1, nbr_loc1, nbr_label2, nbr_loc2)
                            edge_o = Segment(f"O{obs_idx}", self.path[obs_idx])
                            m_next = m.next(edge_m, edge_o, obs=obs_idx)
                            if m_next is not None:
                                self._insert(m_next)
                                if __debug__:
                                    logger.debug(str(m_next))
        if self.max_lattice_width:
            self.lattice[obs_idx].prune(0, self.max_lattice_width, self.expand_now)
        if count == 0:
            if __debug__:
                logger.debug("No active solution found anymore")
            return False
        return True

    def _match_non_emitting_states(self, obs_idx, expand=False):
        """Match sequences of nodes that all refer to the same observation at obs_idx.

        Assumptions:
        This method assumes that the lattice is filled up for both obs_idx and obs_idx + 1.

        :param obs_idx: Index of the first observation used (the second will be obs_idx + 1)
        :return: None
        """
        obs = self.path[obs_idx]
        if obs_idx < len(self.path) - 1:
            obs_next = self.path[obs_idx + 1]
        else:
            obs_next = None
        # The current states are the current observation's states
        if expand:
            cur_lattice = dict((m.key, m) for m in self.lattice[obs_idx].values(0) if not m.stop and m.delayed == self.expand_now)
        else:
            cur_lattice = dict((m.key, m) for m in self.lattice[obs_idx].values(0) if not (m.stop or m.delayed > 0))
        lattice_toinsert = list()
        # The current best states are the next observation's states if you would ignore non-emitting states
        lattice_best = dict((m.shortkey, m)
                            for m in self.lattice[obs_idx + 1].values(0) if not m.stop)
        lattice_ne = set(m.shortkey
                         for m in self.lattice[obs_idx + 1].values(0) if not m.stop and self._skip_ne_states(m))
        # cur_lattice = set(self.lattice[obs_idx].values())
        nb_ne = 0
        prune_thr = None
        while len(cur_lattice) > 0 and nb_ne < self.non_emitting_states_maxnb:
            nb_ne += 1
            if __debug__:
                logger.debug("--- obs {}:{} --- {} - {} ---".format(obs_idx, nb_ne, obs, obs_next))
            cur_lattice = self._match_non_emitting_states_inner(cur_lattice, obs_idx, obs, obs_next, nb_ne,
                                                                lattice_best, lattice_ne)
            if self.max_lattice_width is not None:
                self.lattice[obs_idx].prune(nb_ne, self.max_lattice_width, self.expand_now, prune_thr)
            # Link to next observation
            self._match_non_emitting_states_end(cur_lattice, obs_idx + 1, obs_next,
                                                lattice_best, expand=expand)
            if self.max_lattice_width is not None:
                prune_thr = self.lattice[obs_idx + 1].prune(0, self.max_lattice_width, self.expand_now, None)
        if self.max_lattice_width is not None:
            self.lattice[obs_idx + 1].prune(0, self.max_lattice_width, self.expand_now, None)
        # logger.info('Used {} levels of non-emitting states'.format(nb_ne))
        # for m in lattice_toinsert:
        #     self._insert(m)

    def _node_in_prev_ne(self, m_next, label):
        """Is the given node already visited in the chain of non-emitting states.

        :param m_next:
        :param label: Node label
        :return: True or False
        """
        # for m in itertools.chain(m_next.prev, m_next.prev_other):
        for m in m_next.prev:  # type: BaseMatching
            if m.obs != m_next.obs:
                return False
            assert(m_next.obs_ne != m.obs_ne)
            # print('prev', m.shortkey, 'checking for ', label)
            # if label == m.shortkey:
            if label in m.nodes:
                return True
            if m.obs_ne == 0:
                return False
            if self._node_in_prev_ne(m, label):
                return True
        return False

    @staticmethod
    def _insert_tmp(m_next, lattice):
        if m_next.key in lattice:
            return lattice[m_next.key].update(m_next)
        else:
            lattice[m_next.key] = m_next
            return True

    def _match_non_emitting_states_inner(self, cur_lattice, obs_idx, obs, obs_next, nb_ne,
                                         lattice_best, lattice_ne):
        # cur_lattice_new = dict()
        cur_lattice_new = self.lattice[obs_idx].dict(nb_ne)
        for m in cur_lattice.values():  # type: BaseMatching
            if m.stop or m.delayed != self.expand_now:
                continue
            if m.shortkey in lattice_ne:
                logger.debug(f"Skip non-emitting states from {m.label}, already visited")
                continue
            # == Move to neighbour edge from edge ==
            if m.edge_m.l2 is not None and self.only_edges:
                nbrs = self.map.edges_nbrto((m.edge_m.l1, m.edge_m.l2))
                # print("Neighbours for {}: {}".format(m, nbrs))
                if nbrs is None or len(nbrs) == 0:
                    if __debug__:
                        logger.debug(f"No neighbours found for edge {m.edge_m.label} ({m.label}, non-emitting)")
                    continue
                for nbr_label1, nbr_loc1, nbr_label2, nbr_loc2 in nbrs:
                    if self._node_in_prev_ne(m, nbr_label2):
                        if __debug__:
                            logger.debug(self.matching.repr_static(('x', '{} < node in prev ne'.format(nbr_label2))))
                        continue
                    # === Move to next edge ===
                    if m.edge_m.l2 != nbr_label2 and m.edge_m.l1 != nbr_label2:
                        edge_m = Segment(nbr_label1, nbr_loc1, nbr_label2, nbr_loc2)
                        edge_o = Segment(f"O{obs_idx}", obs, f"O{obs_idx+1}", obs_next)
                        m_next = m.next(edge_m, edge_o, obs=obs_idx, obs_ne=nb_ne)
                        if m_next is not None:
                            if m_next.key in cur_lattice_new:
                                if m_next.shortkey in lattice_best:
                                    if approx_leq(m_next.dist_obs, lattice_best[m_next.shortkey].dist_obs):
                                        cur_lattice_new[m_next.key].update(m_next)
                                    else:
                                        m_next.stop = True
                                        if __debug__ and logger.isEnabledFor(logging.DEBUG):
                                            logger.debug(f"   | Stopped trace: distance larger than best for key {m_next.shortkey}: "
                                                         f"{m_next.dist_obs} > {lattice_best[m_next.shortkey].dist_obs}")
                                else:
                                    cur_lattice_new[m_next.key].update(m_next)
                            else:
                                if m_next.shortkey in lattice_best:
                                    # if m_next.logprob > lattice_best[m_next.shortkey].logprob:
                                    if approx_leq(m_next.dist_obs, lattice_best[m_next.shortkey].dist_obs):
                                        cur_lattice_new[m_next.key] = m_next
                                        # lattice_best[m_next.shortkey] = m_next
                                        # lattice_toinsert.append(m_next)
                                    else:
                                        if __debug__ and logger.isEnabledFor(logging.DEBUG):
                                            logger.debug(f"   | Stopped trace: distance larger than best for key {m_next.shortkey}: "
                                                         f"{m_next.dist_obs} > {lattice_best[m_next.shortkey].dist_obs}")
                                        m_next.stop = True
                                else:
                                    cur_lattice_new[m_next.key] = m_next
                                    # lattice_best[m_next.shortkey] = m_next
                                    # lattice_toinsert.append(m_next)
                            # cur_lattice_new.add(m_next)
                            if __debug__:
                                logger.debug(str(m_next))
                    else:
                        if __debug__:
                            logger.debug(self.matching.repr_static(('x', f'{nbr_label1}-{nbr_label2} < goes back (ne)')))
            # == Move to neighbour node from node==
            if m.edge_m.l2 is None and not self.only_edges:
                cur_node = m.edge_m.l1
                nbrs = self.map.nodes_nbrto(cur_node)
                if nbrs is None:
                    if __debug__:
                        logger.debug(
                            f"No neighbours found for node {cur_node} ({m.label}, non-emitting)")
                    continue
                if __debug__:
                    logger.debug(
                        f"   + Move to {len(nbrs)} neighbours from node {cur_node} ({m.label}, non-emitting)")
                    logger.debug(m.repr_header())
                for nbr_label, nbr_loc in nbrs:
                    # print(f"self._node_in_prev_ne({m.label}, {nbr_label}) = {self._node_in_prev_ne(m, nbr_label)}")
                    if self._node_in_prev_ne(m, nbr_label):
                        if __debug__:
                            logger.debug(self.matching.repr_static(('x', '{} < node in prev ne'.format(nbr_label))))
                        continue
                    # === Move to next node ===
                    if m.edge_m.l1 != nbr_label:
                        edge_m = Segment(nbr_label, nbr_loc)
                        edge_o = Segment(f"O{obs_idx}", obs, f"O{obs_idx+1}", obs_next)
                        m_next = m.next(edge_m, edge_o, obs=obs_idx, obs_ne=nb_ne)
                        if m_next is not None:
                            if m_next.key in cur_lattice_new:
                                cur_lattice_new[m_next.key].update(m_next)
                            else:
                                if m_next.shortkey in lattice_best:
                                    # if m_next.logprob > lattice_best[m_next.shortkey].logprob:
                                    if m_next.dist_obs < lattice_best[m_next.shortkey].dist_obs:
                                        cur_lattice_new[m_next.key] = m_next
                                        lattice_best[m_next.shortkey] = m_next
                                        # lattice_toinsert.append(m_next)
                                    elif __debug__ and logger.isEnabledFor(logging.DEBUG):
                                        m_next.stop = True
                                        cur_lattice_new[m_next.key] = m_next
                                        # lattice_toinsert.append(m_next)
                                else:
                                    cur_lattice_new[m_next.key] = m_next
                                    lattice_best[m_next.shortkey] = m_next
                                    # lattice_toinsert.append(m_next)
                            # cur_lattice_new.add(m_next)
                            if __debug__:
                                logger.debug(str(m_next))
                    else:
                        if __debug__:
                            logger.debug(f"x  | {m.edge_m.l1}-{nbr_label} < self-loop")

        return cur_lattice_new

    def _match_non_emitting_states_end(self, cur_lattice, obs_idx, obs_next,
                                       lattice_best, expand=False):
        for m in cur_lattice.values():  # type: BaseMatching
            if m.stop or m.delayed > self.expand_now:
                continue
            if m.edge_m.l2 is not None:
                # Move to neighbour edge from edge
                nbrs = self.map.edges_nbrto((m.edge_m.l1, m.edge_m.l2))
                # print("Neighbours for {}: {}".format(m, nbrs))
                if nbrs is None or len(nbrs) == 0:
                    if __debug__:
                        logger.debug("No neighbours found for edge {} ({})".format(m.edge_m.label, m.label))
                    continue
                if __debug__:
                    logger.debug(f"   + Move to {len(nbrs)} neighbours from edge {m.edge_m.label} "
                                 f"({m.label}, non-emitting->emitting)")
                    logger.debug(m.repr_header())
                for nbr_label1, nbr_loc1, nbr_label2, nbr_loc2 in nbrs:
                    if self._node_in_prev_ne(m, nbr_label2):
                        if __debug__:
                            logger.debug(self.matching.repr_static(('x', '{} < node in prev ne'.format(nbr_label2))))
                        continue
                    # Move to next edge
                    if m.edge_m.l1 != nbr_label2 and m.edge_m.l2 != nbr_label2:
                        edge_m = Segment(nbr_label1, nbr_loc1, nbr_label2, nbr_loc2)
                        edge_o = Segment(f"O{obs_idx+1}", obs_next)
                        m_next = m.next(edge_m, edge_o, obs=obs_idx)
                        if m_next is not None:
                            if m_next.shortkey in lattice_best:
                                # if m_next.dist_obs < lattice_best[m_next.shortkey].dist_obs:
                                if m_next.logprob > lattice_best[m_next.shortkey].logprob:
                                    lattice_best[m_next.shortkey] = m_next
                                    # lattice_toinsert.append(m_next)
                                    self.lattice[obs_idx].upsert(m_next)
                                elif __debug__ and logger.isEnabledFor(logging.DEBUG):
                                    m_next.stop = True
                                    # lattice_toinsert.append(m_next)
                                    self.lattice[obs_idx].upsert(m_next)
                            else:
                                lattice_best[m_next.shortkey] = m_next
                                # lattice_toinsert.append(m_next)
                                self.lattice[obs_idx].upsert(m_next)
                            if __debug__:
                                logger.debug(str(m_next))
                    else:
                        if __debug__:
                            logger.debug(self.matching.repr_static(('x', '{} < going back'.format(nbr_label2))))
            else:  # m.edge_m.l2 is None:
                # Move to neighbour node from node
                cur_node = m.edge_m.l1
                nbrs = self.map.nodes_nbrto(cur_node)
                # print("Neighbours for {}: {}".format(m, nbrs))
                if nbrs is None:
                    if __debug__:
                        logger.debug("No neighbours found for node {}".format(cur_node, m.label))
                    continue
                if __debug__:
                    logger.debug(f"   + Move to {len(nbrs)} neighbours from node {cur_node} "
                                 f"({m.label}, non-emitting->emitting)")
                    logger.debug(m.repr_header())
                for nbr_label, nbr_loc in nbrs:
                    if self._node_in_prev_ne(m, nbr_label):
                        if __debug__:
                            logger.debug(self.matching.repr_static(('x', '{} < node in prev ne'.format(nbr_label))))
                        continue
                    # Move to next node
                    if m.edge_m.l1 != nbr_label:
                        # edge_m = Segment(m.edge_m.l1, m.edge_m.p1, nbr_label, nbr_loc)
                        edge_m = Segment(nbr_label, nbr_loc)
                        edge_o = Segment(f"O{obs_idx+1}", obs_next)
                        m_next = m.next(edge_m, edge_o, obs=obs_idx)
                        if m_next is not None:
                            if m_next.shortkey in lattice_best:
                                # if m_next.dist_obs < lattice_best[m_next.shortkey].dist_obs:
                                if m_next.logprob > lattice_best[m_next.shortkey].logprob:
                                    lattice_best[m_next.shortkey] = m_next
                                    # lattice_toinsert.append(m_next)
                                    self.lattice[obs_idx].upsert(m_next)
                                elif __debug__ and logger.isEnabledFor(logging.DEBUG):
                                    m_next.stop = True
                                    # lattice_toinsert.append(m_next)
                                    self.lattice[obs_idx].upsert(m_next)
                            else:
                                lattice_best[m_next.shortkey] = m_next
                                # lattice_toinsert.append(m_next)
                                self.lattice[obs_idx].upsert(m_next)
                            if __debug__:
                                logger.debug(str(m_next))
                    else:
                        if __debug__:
                            logger.debug(self.matching.repr_static(('x', '{} < self-loop'.format(nbr_label))))

    def get_matching(self, identifier=None):
        m = None  # type: Optional[BaseMatching]
        if isinstance(identifier, BaseMatching):
            m = identifier
        elif identifier is None:
            col = self.lattice[len(self.lattice) - 1]
            for curm in col.values_all():
                if m is None or curm.logprob > m.logprob:
                    m = curm
        elif type(identifier) is int:
            # If integer, search for the best matching at this index in the lattice
            for cur_m in self.lattice[identifier].values_all():  # type:BaseMatching
                if not cur_m.stop and (m is None or cur_m.logprob > m.logprob):
                    m = cur_m
        elif type(identifier) is str:
            # If string, try to parse identifier
            parts = identifier.split('-')
            idx, ne, key = None, None, None
            if len(parts) == 4:
                nodea, nodeb, idx, ne = [int(part) for part in parts]
                key = (nodea, nodeb, idx, ne)
                col = self.lattice[idx]  # type: LatticeColumn
                col_ne = col.o[ne]
                m = col_ne[key]
            elif len(parts) == 3:
                node, idx, ne = [int(part) for part in parts]
                key = (node, idx, ne)
                col = self.lattice[idx]  # type: LatticeColumn
                col_ne = col.o[ne]
                m = col_ne[key]
            elif len(parts) == 1:
                m = None
                l1 = int(parts[0])
                for l in self.lattice.values():  # type: LatticeColumn
                    for curm in l.values_all():
                        if (curm.edge_m.l1 == l1 or curm.edge_m.l2 == l1) and \
                                (m is None or curm.logprob > m.logprob):
                            m = curm
            else:
                raise AttributeError(f'Unknown string format for matching. '
                                     'Expects <node>-<idx>-<ne> or <node>-<node>-<idx>-<ne>.')

        return m

    def get_matching_path(self, start_m):
        """List of Matching objects that end in the given Matching object."""
        start_m = self.get_matching(start_m)
        return self._build_matching_path(start_m)

    def get_node_path(self, start_m, only_nodes=False):
        """List of node/edge names that end in the given Matching object."""
        path = self.get_matching_path(start_m)
        node_path = [m.shortkey for m in path]
        if only_nodes:
            node_path = self.node_path_to_only_nodes(node_path)
        return node_path

    def node_path_to_only_nodes(self, path):
        """Path of nodes and edges to only nodes.

        :param path: List of node names or edges as (node name, node name)
        :return: List of node names
        """
        nodes = []
        prev_state = path[0]
        if type(prev_state) is tuple:
            nodes.append(prev_state[0])
            nodes.append(prev_state[1])
            prev_node = prev_state[1]
        else:
            nodes.append(prev_state)
            prev_node = prev_state
        for state in path[1:]:
            if state == prev_state:
                continue
            if type(state) is not tuple:
                if state != prev_node:
                    nodes.append(state)
                    prev_node = state
            elif type(state) is tuple:
                if state[0] == prev_node:
                    if state[1] != prev_node:
                        nodes.append(state[1])
                        prev_node = state[1]
                elif state[1] == prev_node:
                    if state[0] != prev_node:
                        nodes.append(state[0])
                        prev_node = state[0]
                else:
                    raise Exception(f"State {state} does not have as previous node {prev_node}")
            else:
                raise Exception(f"Unknown type of state: {state} ({type(state)})")
            prev_state = state
        return nodes

    def _build_matching_path(self, start_m, max_depth=None):
        lattice_best = []
        node_max = start_m
        cur_depth = 0
        if __debug__ and logger.isEnabledFor(logging.DEBUG):
            logger.debug(self.matching.repr_header(stop="             "))
        logger.debug("Start ({}): {}".format(node_max.obs, node_max))
        lattice_best.append(node_max)
        if node_max.is_emitting():
            cur_depth += 1
        # for obs_idx in reversed(range(start_idx)):
        if max_depth is None:
            max_depth = len(self.lattice) + 1
        while cur_depth < max_depth and len(node_max.prev) > 0:
            node_max_last = node_max
            node_max: Optional[BaseMatching] = None
            for prev_m in node_max_last.prev:
                if prev_m is not None and (node_max is None or prev_m.logprob > node_max.logprob):
                    node_max = prev_m
            if node_max is None:
                logger.error("Did not find a matching node for path point at index {}. ".format(node_max_last.obs) +
                             "Stopped building path.")
                break
            logger.debug("Max   ({}): {}".format(node_max.obs, node_max))
            lattice_best.append(node_max)
            if node_max.is_emitting():
                cur_depth += 1
        lattice_best = list(reversed(lattice_best))
        return lattice_best

    def _build_node_path(self, start_idx, unique=True, max_depth=None, last_is_e=False):
        """Build the path from the lattice.

        :param start_idx:
        :param unique:
        :param max_depth:
        :param last_is_e: Last matched lattice node should be an emitting state.
            In case the matching stops early, the longest path can be in between two observations
            and thus be a nonemitting state (which by definition has a lower probability than the
            last emitting state). If this argument is set to true, the longer match is preferred.
        :return:
        """
        node_max = None
        node_max_ne = 0
        if last_is_e:
            for m in self.lattice[start_idx].values_all():  # type:BaseMatching
                if not m.stop and (node_max is None or m.logprob > node_max.logprob):
                    node_max = m
        else:
            for m in self.lattice[start_idx].values_all():  # type:BaseMatching
                if not m.stop and (node_max is None or m.obs_ne > node_max_ne or m.logprob > node_max.logprob):
                    node_max_ne = m.obs_ne
                    node_max = m
        if node_max is None:
            logger.error("Did not find a matching node for path point at index {}".format(start_idx))
            return None

        self.lattice_best = self._build_matching_path(node_max, max_depth)

        node_path = [m.shortkey for m in self.lattice_best]
        if unique:
            self.node_path = []
            prev_node = None
            for node in node_path:
                if node != prev_node:
                    self.node_path.append(node)
                    prev_node = node
        else:
            self.node_path = node_path
        return self.node_path

    def increase_max_lattice_width(self, max_lattice_width, unique=False, tqdm=None):
        self.max_lattice_width = max_lattice_width
        return self.match(self.path, unique=unique, tqdm=tqdm, expand=True)

    def path_bb(self):
        """Get boundig box of matched path (if it exists, otherwise return None)."""
        path = self.path
        plat, plon = islice(zip(*path), 2)
        lat_min, lat_max = min(plat), max(plat)
        lon_min, lon_max = min(plon), max(plon)
        bb = lat_min, lon_min, lat_max, lon_max
        return bb

    def print_lattice(self, file=None, obs_idx=None, obs_ne=0, label_width=None, debug=False):
        if debug:
            xprint = logger.debug
        else:
            if file is None:
                file = sys.stdout
            xprint = lambda arg: print(arg, file=file)
        # print("Lattice:", file=file)
        if obs_idx is not None:
            idxs = [obs_idx]
        else:
            idxs = range(len(self.lattice))
        for idx in idxs:
            if len(self.lattice[idx]) > 0:
                if label_width is None:
                    label_width = 0
                    for m in self.lattice[idx].values(obs_ne):
                        label_width = max(label_width, len(str(m.label)))
                xprint("--- obs {} ---".format(idx))
                xprint(self.matching.repr_header(label_width=label_width))
                for m in sorted(self.lattice[idx].values(obs_ne), key=lambda t: str(t.label)):
                    xprint(m.__str__(label_width=label_width))

    def lattice_dot(self, file=None, precision=None, render=False):
        """Write the lattice as a Graphviz DOT file.

        :param file: File object to print to. Prints to stdout if None.
        :param precision: Precision of (log) probabilities.
        :param render: Try to render the generated Graphviz file.
        """
        if file is None:
            file = sys.stdout
        if precision is None:
            prfmt = ''
        else:
            prfmt = f'.{precision}f'
        print('digraph lattice {', file=file)
        print('\trankdir=LR;', file=file)
        # Vertices
        for idx_ob in range(len(self.lattice)):
            col = self.lattice[idx_ob]
            for idx_ne in range(len(col)):
                ms = col.values(idx_ne)
                if len(ms) == 0:
                    continue
                cnames = [(m.obs_ne, m.cname, m.stop, m.delayed) for m in ms]
                cnames.sort()
                cur_obs_ne = -1
                print('\t{\n\t\trank=same; ', file=file)
                for obs_ne, cname, stop, delayed in cnames:
                    if obs_ne != cur_obs_ne:
                        if cur_obs_ne != -1:
                            print('\t};\n\t{\n\t\trank=same; ', file=file)
                        cur_obs_ne = obs_ne
                    if stop:
                        options = 'label="{} x",color=gray,fontcolor=gray'.format(cname)
                    elif delayed > self.expand_now:
                        options = 'label="{} d{}",color=gray,fontcolor=gray'.format(cname, delayed)
                    elif self.expand_now != 0:
                        options = 'label="{} d{}"'.format(cname, delayed)
                    else:
                        options = 'label="{}  "'.format(cname)
                    print('\t\t{} [{}];'.format(cname, options), file=file)
                print('\t};', file=file)
        # Edges
        for idx_ob in range(len(self.lattice)):
            col = self.lattice[idx_ob]
            for idx_ne in range(len(col)):
                ms = col.values(idx_ne)
                if len(ms) == 0:
                    continue
                for m in ms:
                    for mp in m.prev:
                        if m.stop or m.delayed > self.expand_now:
                            options = ',color=gray,fontcolor=gray'
                        else:
                            options = ''
                        print(f'\t {mp.cname} -> {m.cname} [label="{m.logprob:{prfmt}}"{options}];', file=file)
                    for mp in m.prev_other:
                        if m.stop or m.delayed > self.expand_now:
                            options = ',color=gray,fontcolor=gray'
                        else:
                            options = ''
                        print(f'\t {mp.cname} -> {m.cname} [color=gray,label="{m.logprob:{prfmt}}"{options}];', file=file)
        print('}', file=file)
        if render and file is not None:
            import subprocess as sp
            from pathlib import Path
            from io import TextIOWrapper
            if isinstance(file, Path):
                fn = str(file.canonical())
            elif isinstance(file, TextIOWrapper):
                file.flush()
                fn = file.name
            else:
                fn = str(file)
            cmd = ['dot', '-Tpdf', '-O', fn]
            logger.debug(' '.join(cmd))
            sp.call(cmd)

    def print_lattice_stats(self, file=None, verbose=False):
        if file is None:
            file = sys.stdout
        print("Stats lattice", file=file)
        print("-------------", file=file)
        stats = OrderedDict()
        stats["nbr levels"] = len(self.lattice) if self.lattice else "?"
        total_nodes = 0
        max_nodes = 0
        min_nodes = 9999999
        if self.lattice:
            sizes = []
            for idx in range(len(self.lattice)):
                level = self.lattice[idx].values(0)
                # stats["#nodes[{}]".format(idx)] = len(level)
                sizes.append(len(level))
                total_nodes += len(level)
                if len(level) < min_nodes:
                    min_nodes = len(level)
                if len(level) > max_nodes:
                    max_nodes = len(level)
            stats["nbr lattice"] = total_nodes
            if verbose:
                stats["nbr lattice[level]"] = ", ".join([str(s) for s in sizes])
            stats["avg lattice[level]"] = total_nodes/len(self.lattice)
            stats["min lattice[level]"] = min_nodes
            stats["max lattice[level]"] = max_nodes
        if self.lattice_best and len(self.lattice_best) > 0:
            stats["avg obs distance"] = np.mean([m.dist_obs for m in self.lattice_best])
            stats["last logprob"] = self.lattice_best[-1].logprob
            stats["last length"] = self.lattice_best[-1].length
            stats["last norm logprob"] = self.lattice_best[-1].logprob / self.lattice_best[-1].length
            if verbose:
                stats["best logprob"] = ", ".join(["{:.3f}".format(m.logprob) for m in self.lattice_best])
                stats["best norm logprob"] = \
                    ", ".join(["{:.3f}".format(m.logprob/m.length) for i, m in enumerate(self.lattice_best)])
                stats["best norm prob"] = \
                    ", ".join(["{:.3f}".format(math.exp(m.logprob/m.length)) for i, m in enumerate(self.lattice_best)])
        for key, val in stats.items():
            print("{:<24} : {}".format(key, val), file=file)

    def node_counts(self):
        if self.lattice is None:
            return None
        counts = defaultdict(lambda: 0)
        for level in self.lattice.values():
            for m in level.values_all():
                counts[m.label] += 1
        return counts

    def copy_lastinterface(self, nb_interfaces=1):
        """Copy the current matcher and keep the last interface as the start point.

        This method allows you to perform incremental matching without keeping the entire
        lattice in memory.

        You need to run :meth:`match_incremental` on this object to continue from the existing
        (partial) lattice. Otherwise, if you use :meth:`match`, it will be overwritten.

        Open question, if there is no need to keep track of older lattices, it will probably
        be more efficient to clear the older parts of the interface instead of copying the newer
        parts.

        :param nb_interfaces: Nb of interfaces (columns in lattice) to keep. Default is 1, the last one.
        :return: new Matcher object
        """
        matcher = self.__class__(self.map, obs_noise=self.obs_noise, max_dist_init=self.max_dist_init,
                                 max_dist=self.max_dist, min_prob_norm=self.min_logprob_norm,
                                 non_emitting_states=self.non_emitting_states,
                                 max_lattice_width=self.max_lattice_width, only_edges=self.only_edges,
                                 obs_noise_ne=self.obs_noise_ne, matching=self.matching,
                                 avoid_goingback=self.avoid_goingback,
                                 non_emitting_length_factor=math.exp(self.ne_length_factor_log))
        matcher.lattice = []
        matcher.path = []
        for int_i in range(len(self.lattice) - nb_interfaces, len(self.lattice)):
            matcher.lattice.append(self.lattice[int_i])
            matcher.path.append(self.path[int_i])
        return matcher

    @property
    def path_pred(self):
        """The matched path, both nodes and/or edges (depending on your settings)."""
        return self.node_path

    @property
    def path_pred_onlynodes(self):
        """A list with all the nodes (no edges) the matched path passes through."""
        if self.node_path is None or len(self.node_path) == 0:
            return []
        return self.node_path_to_only_nodes(self.node_path)

    def path_pred_distance(self):
        """Total distance of the matched path."""
        if self.lattice_best is None:
            return None
        if len(self.lattice_best) == 1:
            return 0
        dist = 0
        m_prev = self.lattice_best[0]
        for idx, m in enumerate(self.lattice_best[1:]):
            if m_prev.edge_m.label != m.edge_m.label and m_prev.edge_m.l2 == m.edge_m.l1:
                # Go over the connection between two edges to compute the distance
                cdist = self.map.distance(m_prev.edge_m.pi, m_prev.edge_m.p2)
                cdist += self.map.distance(m_prev.edge_m.p2, m.edge_m.pi)
            else:
                cdist = self.map.distance(m_prev.edge_m.pi, m.edge_m.pi)
            dist += cdist
            m_prev = m
        return dist

    def path_distance(self):
        """Total distance of the observations."""
        if self.lattice_best is None:
            return None
        if len(self.lattice_best) == 1:
            return 0
        dist = 0
        m_prev = self.lattice_best[0]
        for m in self.lattice_best[1:]:
            dist += self.map.distance(m_prev.edge_o.pi, m.edge_o.pi)
            m_prev = m
        return dist

    def path_all_distances(self):
        """Return a list of all distances between observed trace and map.

        One entry for each point in the map and point in the trace that are mapped to each other.
        In case non-emitting nodes are used, extra entries can be present where a point in the trace
        or a point in the map is mapped to a segment.
        """
        path = self.lattice_best
        dists = [m.dist_obs for m in path]
        return dists



#!/usr/bin/env python3
# encoding: utf-8
"""
tests.test_parallelroads
~~~~~~~~~~~~~~~~~~~~~~~~

:author: Wannes Meert
:copyright: Copyright 2018 DTAI, KU Leuven and Sirris.
:license: Apache License, Version 2.0, see LICENSE for details.
"""
import sys
import os
import logging
from pathlib import Path
import leuvenmapmatching as mm
from leuvenmapmatching.map.sqlite import SqliteMap
from leuvenmapmatching.matcher.distance import DistanceMatcher
from leuvenmapmatching.util.dist_euclidean import lines_parallel


logger = mm.logger
directory = None


def create_map1():
    db = SqliteMap("map", use_latlon=False, dir=directory)
    logger.debug(f"Initialized db: {db}")
    db.add_nodes([
        (1, (1, 1)),
        (2, (1, 2.9)),
        (22, (1, 3.0)),
        (3, (2, 2)),
        (33, (2, 2.1)),
        (4, (2, 4)),
        (5, (3, 3)),
        (6, (3, 5))
    ])
    db.add_edges([
        (1, 2), (1, 3),
        (2, 22), (2, 1),
        (22, 2), (22, 33), (22, 4),
        (3, 33), (3, 1), (3, 2), (3, 5),
        (33, 3),
        (4, 22), (4, 33), (4, 5), (4, 6),
        (5, 3), (5, 4), (5, 6),
        (6, 4), (6, 5)
    ])
    logger.debug(f"Filled db: {db}")
    return db


def create_path1():
    return [(0.9, 2.5), (1.1, 2.75), (1.25, 2.6), (1.4, 2.5), (1.5, 2.4), (1.6, 2.5), (1.4, 2.7), (1.2, 2.9), (1.1, 3.0), (1.3, 3.2)]


def test_parallel():
    result = lines_parallel((1, 2.9), (2, 2), (1, 3.0), (2, 2.1), d=0.1)
    assert result is True


def test_bb1():
    mapdb = create_map1()

    if directory:
        from leuvenmapmatching import visualization as mmviz
        mmviz.plot_map(mapdb,
                       show_labels=True, show_graph=True,
                       filename=str(directory / "test_bb.png"))

    mapdb.connect_parallelroads()

    assert mapdb.size() == 8
    coord = mapdb.node_coordinates(2)
    assert coord == (1, 2.9)

    nodes = mapdb.all_nodes(bb=[0.5, 2.5, 1.5, 3.5])
    node_ids = set([nid for nid, _ in nodes])
    assert node_ids == {2, 22}

    edges = mapdb.all_edges(bb=[0.5, 2.5, 1.5, 3.5])
    edge_tuples = set((nid1, nid2) for nid1, _, nid2, _ in edges)
    assert edge_tuples == {(1, 2), (2, 1), (3, 2), (2, 22), (22, 2), (22, 33), (22, 4), (4, 22)}

    nodes = mapdb.nodes_nbrto(2)
    node_ids = set([nid for nid, _ in nodes])
    assert node_ids == {1, 22}

    edges = mapdb.edges_nbrto((1, 2))
    edge_tuples = set((nid1, nid2) for nid1, _, nid2, _ in edges)
    assert edge_tuples == {(2, 22), (2, 1)}

    edges = mapdb.edges_nbrto((3, 2))
    edge_tuples = set((nid1, nid2) for nid1, _, nid2, _ in edges)
    assert edge_tuples == {(22, 33), (2, 22), (2, 1)}


def test_merge1():
    mapdb = create_map1()
    mapdb.connect_parallelroads()

    if directory:
        from leuvenmapmatching import visualization as mmviz
        mmviz.plot_map(mapdb,
                       show_labels=True, show_graph=True,
                       filename=str(directory / "test_parallel_merge.png"))


def test_path1():
    mapdb = create_map1()
    mapdb.connect_parallelroads()
    path = create_path1()
    states_sol = [(1, 2), (2, 22), (22, 33), (22, 33), (22, 33), (3, 2), (3, 2), (3, 2), (2, 22), (22, 4)]

    matcher = DistanceMatcher(mapdb, max_dist_init=0.2,
                              obs_noise=0.5, obs_noise_ne=2, dist_noise=0.5,
                              non_emitting_states=True)
    states, _ = matcher.match(path)

    if directory:
        from leuvenmapmatching import visualization as mmviz
        mmviz.plot_map(mapdb, matcher=matcher,
                       show_labels=True, show_graph=True, show_matching=True,
                       filename=str(directory / "test_parallel_merge.png"))
    assert states == states_sol, f"Unexpected states: {states}"


if __name__ == "__main__":
    logger.setLevel(logging.DEBUG)
    logger.addHandler(logging.StreamHandler(sys.stdout))
    directory = Path(os.environ.get('TESTDIR', Path(__file__).parent))
    print(f"Saving files to {directory}")
    # test_parallel()
    # test_merge1()
    # test_path1()
    test_bb1()



#!/usr/bin/env python3
# encoding: utf-8
"""
setup.py
~~~~~~~~

:author: Wannes Meert
:copyright: Copyright 2017-2021 DTAI, KU Leuven and Sirris.
:license: Apache License, Version 2.0, see LICENSE for details.
"""
from setuptools import setup, find_packages
import re
import os

here = os.path.abspath(os.path.dirname(__file__))


with open(os.path.join('leuvenmapmatching', '__init__.py'), 'r') as fd:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
                        fd.read(), re.MULTILINE).group(1)
if not version:
    raise RuntimeError('Cannot find version information')


setup(
    name='leuvenmapmatching',
    version=version,
    packages=find_packages(),
    author='Wannes Meert',
    author_email='wannes.meert@cs.kuleuven.be',
    url='https://github.com/wannesm/LeuvenMapMatching',
    description='Match a trace of GPS positions to a locations and streets on a map',
    python_requires='>=3.6',
    license='Apache 2.0',
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3'
    ],
    keywords='map matching',
)



# encoding: utf-8
"""
leuvenmapmatching.util.evaluation
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Methods to help set up and evaluate experiments.

:author: Wannes Meert
:copyright: Copyright 2018 DTAI, KU Leuven and Sirris.
:license: Apache License, Version 2.0, see LICENSE for details.
"""
import logging
from dtaidistance.alignment import needleman_wunsch, best_alignment

from . import dist_latlon
MYPY = False
if MYPY:
    from ..map.base import BaseMap
    from typing import List, Tuple, Optional, Callable


logger = logging.getLogger("be.kuleuven.cs.dtai.mapmatching")


def route_mismatch_factor(map_con, path_pred, path_grnd, window=None, dist_fn=None, keep_mismatches=False):
    # type: (BaseMap, List[int], List[int], Optional[int], Optional[Callable], bool) -> Tuple[float, float, float, float, List[Tuple[int, int]], float, float]
    """Evaluation method from Newson and Krumm (2009).

    :math:`f = \frac{d_{-} + d_{+}}{d_0}`

    With :math:`d_{-}` the length that is erroneously subtracted,
    :math:`d_{+}` the length that is erroneously added, and :math:`d_0` the
    distance of the correct route.

    This function only supports connected states (thus not switching between states
    that are not connected (e.g. parallel roads).

    Also computes the Accuracy by Number (AN) and Accuracy by Length (AL) metrics from
    Zheng et al. (2009).
    """
    if dist_fn is None:
        dist_fn = dist_latlon.distance
    _, matrix = needleman_wunsch(path_pred, path_grnd, window=window)
    print(matrix[:10, :10])
    algn, _, _ = best_alignment(matrix)
    print(algn[:10])
    d_plus = 0  # length erroneously added
    d_min = 0  # length erroneously subtracted
    d_zero = 0  # length of correct route
    cnt_matches = 0  # number of perfect matches
    cnt_mismatches = 0
    mismatches = [] if keep_mismatches else None

    prev_grnd_pi = None
    for pred_pi, grnd_pi in algn:
        pred_p = path_pred[pred_pi]
        grnd_p = path_grnd[grnd_pi]
        grnd_d = map_con.path_dist(grnd_p)
        d_zero += grnd_d
        if pred_p == grnd_p:
            cnt_matches += 1
        else:
            # print(f"Mismatch: {pred_p} != {grnd_p}")
            cnt_mismatches += 1
            pred_d = map_con.path_dist(pred_p)
            d_plus += pred_d
            d_min += grnd_d
            if keep_mismatches:
                mismatches.append((pred_p, grnd_p))
        prev_grnd_pi = grnd_pi

    factor = (d_min + d_plus) / d_zero
    an = cnt_matches / len(path_grnd)
    al = (d_zero - d_min) / d_zero
    return factor, cnt_matches, cnt_mismatches, d_zero, mismatches, an, al



# encoding: utf-8
"""
leuvenmapmatching.matcher.newsonkrumm
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Methods similar to Newson Krumm 2009 for comparison purposes.

    P. Newson and J. Krumm. Hidden markov map matching through noise and sparseness.
    In Proceedings of the 17th ACM SIGSPATIAL international conference on advances
    in geographic information systems, pages 336–343. ACM, 2009.



:author: Wannes Meert
:copyright: Copyright 2018 DTAI, KU Leuven and Sirris.
:license: Apache License, Version 2.0, see LICENSE for details.
"""
from scipy.stats import norm
import math
import logging

from .base import BaseMatching, BaseMatcher

logger = logging.getLogger("be.kuleuven.cs.dtai.mapmatching")


class NewsonKrummMatching(BaseMatching):
    __slots__ = ['d_s', 'd_o', 'lpe', 'lpt']  # Additional fields

    def __init__(self, *args, d_s=1.0, d_o=1.0, lpe=0.0, lpt=0.0, **kwargs):
        """

        :param args: Arguments for BaseMatching
        :param d_s: Distance between two (interpolated) states
        :param d_o: Distance between two (interpolated) observations
        :param lpe: Log probability of emission
        :param lpt: Log probablity of transition
        :param kwargs: Arguments for BaseMatching
        """
        super().__init__(*args, **kwargs)
        self.d_o: float = d_o
        self.d_s: float = d_s
        self.lpe: float = lpe
        self.lpt: float = lpt

    def _update_inner(self, m_other):
        # type: (NewsonKrummMatching, NewsonKrummMatching) -> None
        super()._update_inner(m_other)
        self.d_s = m_other.d_s
        self.d_o = m_other.d_o
        self.lpe = m_other.lpe
        self.lpt = m_other.lpt

    @staticmethod
    def repr_header(label_width=None, stop=""):
        res = BaseMatching.repr_header(label_width)
        res += f" {'dt(o)':<6} | {'dt(s)':<6} |"
        if logger.isEnabledFor(logging.DEBUG):
            res += f" {'lg(Pr-t)':<9} | {'lg(Pr-e)':<9} |"
        return res

    def __str__(self, label_width=None):
        res = super().__str__(label_width)
        res += f" {self.d_o:>6.2f} | {self.d_s:>6.2f} |"
        if logger.isEnabledFor(logging.DEBUG):
            res += f" {self.lpt:>9.2f} | {self.lpe:>9.2f} |"
        return res


class NewsonKrummMatcher(BaseMatcher):
    """
    Take distance between observations vs states into account. Based on the
    method presented in:

        P. Newson and J. Krumm. Hidden markov map matching through noise and sparseness.
        In Proceedings of the 17th ACM SIGSPATIAL international conference on advances
        in geographic information systems, pages 336–343. ACM, 2009.

    Two important differences:

    * Newson and Krumm use shortest path to handle situations where the distances between
      observations are larger than distances between nodes in the graph. The LeuvenMapMatching
      toolbox uses non-emitting states to handle this. We thus do not implement the shortest
      path algorithm in this class.
    * Transition and emission probability are transformed from densities to probababilities by
      taking the 1 - CDF instead of the PDF.


    Newson and Krumm defaults:

    - max_dist = 200 m
    - obs_noise = 4.07 m
    - beta = 1/6
    - only_edges = True
    """

    def __init__(self, *args, **kwargs):
        """

        :param beta: Default is 1/6
        :param beta_ne: Default is beta
        :param args: Arguments for BaseMatcher
        :param kwargs: Arguments for BaseMatcher
        """

        if not kwargs.get("only_edges", True):
            logger.warning("The MatcherDistance method only works on edges as states. Nodes have been disabled.")
        kwargs["only_edges"] = True
        if "matching" not in kwargs:
            kwargs["matching"] = NewsonKrummMatching
        super().__init__(*args, **kwargs)

        # if not use_original, the following value for beta gives a prob of 0.5 at dist=x_half:
        # beta = np.sqrt(np.power(x_half, 2) / (np.log(2)*2))
        self.beta = kwargs.get('beta', 1/6)
        self.beta_ne = kwargs.get('beta_ne', self.beta)

        self.obs_noise_dist = norm(scale=self.obs_noise)
        self.obs_noise_dist_ne = norm(scale=self.obs_noise_ne)
        self.ne_thr = 1.25

    def logprob_trans(self, prev_m: NewsonKrummMatching, edge_m, edge_o,
                      is_prev_ne=False, is_next_ne=False):
        """Transition probability.

        Main difference with Newson and Krumm: we know all points are connected thus do not compute the
        shortest path but the distance between two points.

        Original PDF:
        p(dt) = 1 / beta * e^(-dt / beta)
        with beta = 1/6

        Transformed to probability:
        P(dt) = p(d > dt) = e^(-dt / beta)

        :param prev_m:
        :param edge_m:
        :param edge_o:
        :param is_prev_ne:
        :param is_next_ne:
        :return:
        """
        d_z = self.map.distance(prev_m.edge_o.pi, edge_o.pi)
        if prev_m.edge_m.label == edge_m.label:
            d_x = self.map.distance(prev_m.edge_m.pi, edge_m.pi)
        else:
            d_x = self.map.distance(prev_m.edge_m.pi, prev_m.edge_m.p2) + self.map.distance(prev_m.edge_m.p2, edge_m.pi)
        d_t = abs(d_z - d_x)
        # p_dt = 1 / beta * math.exp(-d_t / beta)
        if is_prev_ne or is_next_ne:
            beta = self.beta_ne
        else:
            beta = self.beta
        # icp_dt = math.exp(-d_t / beta)
        # try:
        #     licp_dt = math.log(icp_dt)
        # except ValueError:
        #     licp_dt = float('-inf')
        licp_dt = -d_t / beta
        props = {
            'd_o': d_z,
            'd_s': d_x,
            'lpt': licp_dt
        }
        return licp_dt, props

    def logprob_obs(self, dist, prev_m, new_edge_m, new_edge_o, is_ne=False):
        """Emission probability for emitting states.

        Original pdf:
        p(d) = N(0, sigma)
        with sigma = 4.07m

        Transformed to probability:
        P(d) = 2 * (1 - p(d > D)) = 2 * (1 - cdf)

        """
        if is_ne:
            result = 2 * (1 - self.obs_noise_dist_ne.cdf(dist))
        else:
            result = 2 * (1 - self.obs_noise_dist.cdf(dist))
        try:
            result = math.log(result)
        except ValueError:
            result = -float("inf")
        props = {
            'lpe': result
        }
        return result, props
